#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.01.17 13:00:00                  #
# ================================================== #

import os
from typing import Optional

from pygpt_net.core.tabs.tab import Tab
from pygpt_net.core.events import Event, BaseEvent
from pygpt_net.item.ctx import CtxItem
from pygpt_net.utils import trans


class Audio:
    def __init__(self, window=None):
        """
        Audio/voice controller

        :param window: Window instance
        """
        self.window = window
        self.input_allowed_tabs = [
            Tab.TAB_NOTEPAD,
            Tab.TAB_CHAT,
            Tab.TAB_TOOL_CALENDAR,
        ]

    def setup(self):
        """Setup controller"""
        self.update()
        if self.window.core.config.get("audio.input.continuous", False):
            self.window.ui.plugin_addon['audio.input.btn'].continuous.setChecked(True)

    def toggle_input(
            self,
            state: bool,
            btn: bool = True
    ):
        """
        Toggle audio input

        :param state: True to enable, False to disable
        :param btn: True if called from button
        """
        self.window.dispatch(
            Event(Event.AUDIO_INPUT_TOGGLE, {
                "value": state,
            })
        )

    def toggle_output(self):
        """Toggle audio output"""
        if self.window.controller.plugins.is_enabled('audio_output'):
            self.disable_output()
        else:
            self.enable_output()

    def toggle_continuous(self, state: bool):
        """
        Toggle continuous audio input

        :param state: True to enable, False to disable
        """
        if state:
            self.window.core.config.set("audio.input.continuous", True)
        else:
            self.window.core.config.set("audio.input.continuous", False)
        self.window.core.config.save()

    def on_tab_changed(self, tab: Tab):
        """
        On tab changed event

        :param tab: Tab instance (current tab)
        """
        # input button visibility
        if self.is_input_enabled():
            if tab.type in self.input_allowed_tabs:
                self.handle_audio_input(True)  # show btn
            else:
                self.handle_audio_input(False) # hide btn

    def enable_output(self):
        """Enable audio output"""
        self.toggle_output_icon(True)
        self.window.controller.plugins.enable('audio_output')
        self.window.core.config.save()
        self.update()

    def disable_output(self):
        """Disable audio output"""
        self.toggle_output_icon(False)
        self.window.controller.plugins.disable('audio_output')
        self.window.core.config.save()
        self.update()

    def enable_input(self):
        """Enable audio input"""
        self.window.controller.plugins.enable('audio_input')
        self.window.core.config.save()
        self.update()

    def disable_input(self, update: bool = True):
        """
        Disable audio input

        :param update: True to update menu and listeners
        """
        self.window.controller.plugins.disable('audio_input')
        self.window.core.config.save()
        if update:
            self.update()

    def stop_input(self):
        """Stop audio input"""
        self.window.dispatch(
            Event(Event.AUDIO_INPUT_STOP, {
                "value": True,
            }), all=True)

    def stop_output(self):
        """Stop audio output"""
        self.window.dispatch(
            Event(Event.AUDIO_OUTPUT_STOP, {
                "value": True,
            }), all=True)

    def update(self):
        """Update UI and listeners"""
        self.update_listeners()
        self.update_menu()

    def is_output_enabled(self) -> bool:
        """
        Check if any audio output is enabled

        :return: True if enabled
        """
        if self.window.controller.plugins.is_enabled('audio_output'):
            return True
        return False

    def is_input_enabled(self) -> bool:
        """
        Check if any audio input is enabled

        :return: True if enabled
        """
        if self.window.controller.plugins.is_enabled('audio_input'):
            return True
        return False

    def update_listeners(self):
        """Update audio listeners"""
        is_output = False
        if self.window.controller.plugins.is_enabled('audio_output'):
            is_output = True
        if not is_output:
            self.stop_output()

        if not self.window.controller.plugins.is_enabled('audio_input'):
            self.toggle_input(False)
            self.stop_input()
            if self.window.ui.plugin_addon['audio.input'].btn_toggle.isChecked():
                self.window.ui.plugin_addon['audio.input'].btn_toggle.setChecked(False)

    def update_menu(self):
        """Update audio menu"""
        if self.window.controller.plugins.is_enabled('audio_output'):
            self.window.ui.menu['audio.output'].setChecked(True)
        else:
            self.window.ui.menu['audio.output'].setChecked(False)

        if self.window.controller.plugins.is_enabled('audio_input'):
            self.window.ui.menu['audio.input'].setChecked(True)
        else:
            self.window.ui.menu['audio.input'].setChecked(False)

        if self.window.controller.plugins.is_enabled('voice_control'):
            self.window.ui.menu['audio.control.plugin'].setChecked(True)
        else:
            self.window.ui.menu['audio.control.plugin'].setChecked(False)

        if self.window.controller.access.voice.is_voice_control_enabled():
            self.window.ui.menu['audio.control.global'].setChecked(True)
        else:
            self.window.ui.menu['audio.control.global'].setChecked(False)

    def read_text(
            self,
            text: str,
            cache_file: Optional[str] = None
    ):
        """
        Read text using audio output plugins

        :param text: text to read
        :param cache_file: cache file to save
        """
        if text is None or text.strip() == "":
            return

        ctx = CtxItem()
        ctx.output = text
        all = True  # to all plugins (even if disabled)
        event = Event(Event.AUDIO_READ_TEXT)
        event.ctx = ctx
        event.data = {
            "text": text,
            'cache_file': cache_file,
        }
        self.window.dispatch(event, all=all)

    def play_chat_audio(self, path: str):
        """
        Play audio file (chat multimodal response)

        :param path: audio file path
        """
        if not self.is_output_enabled():
            return
        self.play_audio(path)

    def play_audio(self, path: str):
        """
        Play audio file

        :param path: audio file path
        """
        ctx = CtxItem()
        event = Event(Event.AUDIO_PLAYBACK)
        event.ctx = ctx
        event.data = {
            'audio_file': path,
        }
        self.window.dispatch(event, all=True)

    def stop_audio(self):
        """Stop audio playback"""
        ctx = CtxItem()
        event = Event(Event.AUDIO_OUTPUT_STOP)
        event.ctx = ctx
        event.data = {}
        self.window.dispatch(event, all=True)

    def play_sound(self, filename: str):
        """
        Play sound

        :param filename: sound file name
        """
        path = os.path.join(self.window.core.config.get_app_path(), "data", "audio", filename)
        if path:
            self.play_audio(path)

    def play_event(
            self,
            text: str,
            event: Optional[BaseEvent] = None
    ):
        """
        Play event (read text or play cached audio file)

        :param text: text to read
        :param event: event
        """
        use_cache = True
        # event is required to use cache
        if event is None:
            use_cache = False
        else:
            # check if cache is allowed for this event
            if self.window.core.access.voice.cache_disabled(event.name):
                use_cache = False

            # check if not disabled in config
            if not self.window.core.config.get("access.audio.use_cache"):
                use_cache = False

        if text is None or text.strip() == "":
            return

        if use_cache:
            lang = self.window.core.config.get("lang")
            cache_dir = os.path.join(self.window.core.config.get_user_path(), "cache", "audio", lang)
            if not os.path.exists(cache_dir):
                os.makedirs(cache_dir, exist_ok=True)
            cache_file = os.path.join(str(cache_dir), event.name + ".wav")
            # print("Cache file: {}".format(cache_file))
            if os.path.exists(cache_file):
                # print("Using cached file: {}".format(cache_file))
                self.play_audio(cache_file)
            else:
                self.read_text(text, cache_file)
        else:
            self.read_text(text)  # without cache

    def clear_cache(self, force: bool = False):
        """
        Clear audio cache

        :param force: True to force clear
        """
        if not force:
            self.window.ui.dialogs.confirm(
                type='audio.cache.clear',
                id=0,
                msg=trans("audio.cache.clear.confirm"),
            )
            return

        cache_dir = os.path.join(self.window.core.config.get_user_path(), "cache", "audio")
        if os.path.exists(cache_dir):
            import shutil
            shutil.rmtree(cache_dir)
        self.window.ui.dialogs.alert(trans("audio.cache.clear.success"))

    def toggle_output_icon(self, state: bool):
        """
        Toggle input icon

        :param state: True to enable, False to disable
        """
        if state:
            self.window.ui.nodes['icon.audio.output'].set_icon(":/icons/volume.svg")
        else:
            self.window.ui.nodes['icon.audio.output'].set_icon(":/icons/mute.svg")

    def toggle_input_icon(self, state: bool):
        """
        Toggle input icon

        :param state: True to enable, False to disable
        """
        if state:
            self.window.ui.nodes['icon.audio.input'].set_icon(":/icons/mic.svg")
        else:
            self.window.ui.nodes['icon.audio.input'].set_icon(":/icons/mic_off.svg")

    def on_begin(self, text: str):
        """
        On audio playback init

        :param text: text to play
        """
        self.window.update_status(trans("status.audio.start"))

    def on_play(self, event: str):
        """
        On audio playback start

        :param event: event name
        """
        if event == Event.AUDIO_READ_TEXT:
            self.window.update_status("")

    def on_stop(self):
        """
        On audio playback stopped (force)
        """
        self.window.update_status(trans("status.audio.stopped"))

    def is_playing(self) -> bool:
        """
        Check if any audio is playing

        :return: True if playing
        """
        from pygame import mixer
        try:
            mixer.init()
            return mixer.get_busy()
        except Exception as e:
            pass
        return False

    def handle_audio_input(
            self,
            is_enabled: bool
    ):
        """
        Handle audio input UI

        :param is_enabled: enable/disable audio input
        """
        # get advanced audio input option
        is_advanced = False
        data = {
            'name': 'audio.input.advanced',
            'value': is_advanced,
        }
        event = Event(Event.PLUGIN_OPTION_GET, data)
        self.window.dispatch(event)
        if 'value' in event.data:
            is_advanced = event.data['value']
        if is_enabled:
            # show/hide extra options
            tab = self.window.controller.ui.tabs.get_current_tab()
            if tab.type == Tab.TAB_NOTEPAD:
                self.window.ui.plugin_addon['audio.input.btn'].notepad_footer.setVisible(True)
            else:
                self.window.ui.plugin_addon['audio.input.btn'].notepad_footer.setVisible(False)
            if is_advanced:
                self.window.ui.plugin_addon['audio.input.btn'].setVisible(False)
                self.window.ui.plugin_addon['audio.input'].setVisible(True)
            else:
                self.window.ui.plugin_addon['audio.input.btn'].setVisible(True)  # simple recording
                self.window.ui.plugin_addon['audio.input'].setVisible(False)  # advanced recording
            self.toggle_input_icon(True)
        else:
            self.window.ui.plugin_addon['audio.input.btn'].setVisible(False)  # simple recording
            self.window.ui.plugin_addon['audio.input'].setVisible(False)  # advanced recording
            self.toggle_input_icon(False)

    def handle_audio_output(self, is_enabled: bool):
        """
        Handle audio output UI

        :param is_enabled: enable/disable audio output
        """
        if is_enabled:
            self.toggle_output_icon(True)
            # self.window.ui.plugin_addon['audio.output'].setVisible(True)
        else:
            self.window.ui.plugin_addon['audio.output'].setVisible(False)
            self.toggle_output_icon(False)




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.14 08:00:00                  #
# ================================================== #

import os
from typing import Optional, Dict, List

from PySide6.QtWidgets import QApplication

from pygpt_net.item.assistant import AssistantItem
from pygpt_net.item.attachment import AttachmentItem
from pygpt_net.item.ctx import CtxItem
from pygpt_net.utils import trans


class Files:
    def __init__(self, window=None):
        """
        Assistants files controller

        :param window: Window instance
        """
        self.window = window

    def update(self):
        """Update assistants files list"""
        self.update_list()

    def select(self, idx: int):
        """
        Select file

        :param idx: index of file
        """
        id = self.window.core.config.get('assistant')
        if id is None or id == "":
            return
        assistant = self.window.core.assistants.get_by_id(id)
        if assistant is None:
            return

        # get file by list index
        thread_id = self.window.core.config.get('assistant_thread')
        file_id = self.window.core.assistants.files.get_file_id_by_idx(idx, assistant.vector_store, thread_id)
        self.window.core.assistants.current_file = file_id

    def count_upload(
            self,
            attachments: Dict[str, AttachmentItem]
    ) -> int:
        """
        Count attachments for upload

        :param attachments: attachments dict
        :return: number of files need to be uploaded
        """
        num = 0
        for id in list(attachments):
            attachment = attachments[id]
            if not attachment.send:
                num += 1  # increment uploaded files counter if file is not uploaded yet
        return num

    def import_files(self, store_id: Optional[str] = None):
        """
        Import assistant files from API

        :param store_id: store ID
        """
        # run asynchronous
        self.window.update_status("Importing files...please wait...")
        self.window.core.gpt.assistants.importer.import_files(store_id)

    def download(self, idx: int):
        """
        Download file

        :param idx: selected attachment index
        """
        id = self.window.core.config.get('assistant')
        if id is None or id == "":
            return
        assistant = self.window.core.assistants.get_by_id(id)
        if assistant is None:
            return

        # get file by list index
        thread_id = self.window.core.config.get('assistant_thread')
        file_id = self.window.core.assistants.files.get_file_id_by_idx(idx, assistant.vector_store, thread_id)

        # download file
        self.window.controller.attachment.download(file_id)

    def rename(self, idx: int):
        """
        Rename file

        :param idx: selected attachment index
        """
        id = self.window.core.config.get('assistant')
        if id is None or id == "":
            return
        assistant = self.window.core.assistants.get_by_id(id)
        if assistant is None:
            return

        thread_id = self.window.core.config.get('assistant_thread')

        # get file by list index
        file = self.window.core.assistants.files.get_file_by_idx(idx, assistant.vector_store, thread_id)
        if file is None:
            return

        # set dialog and show
        self.window.ui.dialog['rename'].id = 'attachment_uploaded'
        self.window.ui.dialog['rename'].input.setText(file.name)
        self.window.ui.dialog['rename'].current = file.record_id
        self.window.ui.dialog['rename'].show()
        self.update()

    def rename_close(self):
        """Close rename dialog"""
        # close rename dialog and update attachments list
        self.window.ui.dialog['rename'].close()
        self.update()

    def update_name(self, record_id: str, name: str):
        """
        Rename uploaded remote file name

        :param record_id: file record ID
        :param name: new name
        """
        self.window.core.assistants.files.rename(
            record_id,
            name,
        )
        self.rename_close()

    def clear(self, force: bool = False):
        """
        Delete all files

        :param force: force clear files
        """
        if not force:
            self.window.ui.dialogs.confirm(
                type='attachments.uploaded.clear',
                id=-1,
                msg=trans('attachments_uploaded.clear.confirm'),
            )
            return

        id = self.window.core.config.get('assistant')
        if id is None or id == "":
            return

        # delete all files in current thread
        if self.window.core.assistants.has(id):
            assistant = self.window.core.assistants.get_by_id(id)
            thread_id = self.window.core.config.get('assistant_thread')
            items = self.window.core.assistants.files.get_by_store_or_thread(assistant.vector_store, thread_id)
            self.window.update_status(trans('status.sending'))
            QApplication.processEvents()

            for id in list(items.keys()):
                file = items[id]
                try:
                    self.window.core.assistants.files.delete(file)  # delete from DB, API and vector stores
                except Exception as e:
                    self.window.update_status(trans('status.error'))
                    self.window.ui.dialogs.alert(e)

            # update store status
            if assistant.vector_store:
                self.window.controller.assistant.store.refresh_by_store_id(assistant.vector_store)

            self.window.update_status(trans('status.deleted'))

        self.update()

    def delete(self, idx: int, force: bool = False):
        """
        Delete file

        :param idx: file idx
        :param force: force delete without confirmation
        """
        if not force:
            self.window.ui.dialogs.confirm(
                type='attachments.uploaded.delete',
                id=idx,
                msg=trans('attachments_uploaded.delete.confirm'),
            )
            return

        # get current assistant
        id = self.window.core.config.get('assistant')
        if id is None or id == "":
            return
        assistant = self.window.core.assistants.get_by_id(id)
        if assistant is None:
            return

        # get file by list index
        thread_id = self.window.core.config.get('assistant_thread')
        file = self.window.core.assistants.files.get_file_by_idx(idx, assistant.vector_store, thread_id)
        if file is None:
            return

        # delete file in API
        self.window.update_status(trans('status.sending'))
        QApplication.processEvents()
        try:
            self.window.core.assistants.files.delete(file)  # delete from DB, API and vector stores

            # update store status
            if assistant.vector_store:
                self.window.controller.assistant.store.refresh_by_store_id(assistant.vector_store)

            self.window.update_status(trans('status.deleted'))
        except Exception as e:
            self.window.core.debug.log(e)
            self.window.ui.dialogs.alert(e)
            self.window.update_status(trans('status.error'))

        self.update()

    def clear_attachments(self, assistant: AssistantItem):
        """
        Clear attachments

        :param assistant: assistant object
        """
        assistant.clear_attachments()
        self.window.core.assistants.save()
        self.update()

    def upload(
            self,
            mode: str,
            attachments: Dict[str, AttachmentItem]
    ) -> int:
        """
        Upload attachments to assistant

        :param mode: work mode
        :param attachments: attachments dict
        :return: number of uploaded files
        """
        # get current chosen assistant
        assistant_id = self.window.core.config.get('assistant')
        thread_id = self.window.core.config.get('assistant_thread')
        if assistant_id is None:
            return 0

        assistant = self.window.core.assistants.get_by_id(assistant_id)
        num = 0
        # loop on attachments
        for id in list(attachments):
            attachment = attachments[id]
            old_id = attachment.id  # tmp id

            # check if not already uploaded (ignore already uploaded files)
            if not attachment.send:
                msg = "Uploading file: {}".format(attachment.path)
                self.window.core.debug.info(msg, False)
                # check if file exists
                if not os.path.exists(attachment.path):
                    continue

                # upload local attachment file and get new ID (file_id)
                new_id = self.window.core.gpt.store.upload(
                    attachment.path,
                )
                if new_id is not None:
                    # mark as already uploaded
                    attachment.send = True
                    attachment.id = new_id
                    attachment.remote = new_id

                    """
                    if assistant.vector_store is None or assistant.vector_store == "":
                        assistant.vector_store = self.window.core.gpt.store.create_store(
                            "thread-" + thread_id,
                        )
                    """

                    # add to vector store if defined in assistant, otherwise file will be added to thread store
                    if assistant.vector_store:
                        self.window.core.gpt.store.add_file(
                            assistant.vector_store,
                            new_id,
                        )

                    # replace old ID with new one
                    self.window.core.attachments.replace_id(
                        mode,
                        old_id,
                        attachment,
                    )

                    self.window.core.assistants.files.create(
                        assistant,
                        thread_id,
                        new_id,
                        attachment.name,
                        attachment.path,
                        os.path.getsize(attachment.path),
                    )

                    # update assistant attachments list
                    self.window.core.assistants.replace_attachment(
                        assistant,
                        attachment,
                        old_id,
                        new_id,
                    )
                    attachment.consumed = True
                    num += 1  # increment uploaded files counter

        # update assistants list
        self.window.core.assistants.save()

        # update attachments UI
        self.window.controller.attachment.update()

        # update uploaded list
        if num > 0:
            # update store status
            if assistant.vector_store:
                self.window.controller.assistant.store.refresh_by_store_id(assistant.vector_store)

            self.update_list()  # update uploaded list UI

        return num

    def append(
            self,
            assistant: AssistantItem,
            attachment: AttachmentItem
    ):
        """
        Append attachment to assistant

        :param assistant: assistant object
        :param attachment: attachment object
        """
        # get current chosen assistant
        assistant.add_attachment(attachment)  # append attachment
        self.window.core.assistants.save()  # save assistants

    def update_list(self):
        """Update uploaded files list"""
        assistant_id = self.window.core.config.get('assistant')
        thread_id = self.window.core.config.get('assistant_thread')
        if assistant_id is None or assistant_id == "":
            return
        assistant = self.window.core.assistants.get_by_id(assistant_id)
        if assistant is None:
            return
        items = self.window.core.assistants.files.get_by_store_or_thread(assistant.vector_store, thread_id)
        self.window.ui.chat.input.attachments_uploaded.update(items)
        self.update_tab()

    def update_tab(self):
        """Update tab label (attachments uploaded)"""
        assistant_id = self.window.core.config.get('assistant')
        thread_id = self.window.core.config.get('assistant_thread')
        if assistant_id is None or assistant_id == "":
            self.window.ui.tabs['input'].setTabText(2, trans('attachments_uploaded.tab'))
            return
        assistant = self.window.core.assistants.get_by_id(assistant_id)
        if assistant is None:
            return  # no assistant
        num_files = self.window.core.assistants.files.count_by_store_or_thread(assistant.vector_store, thread_id)
        suffix = ''
        # append num of files
        if num_files > 0:
            suffix = f' ({num_files})'
        self.window.ui.tabs['input'].setTabText(2, trans('attachments_uploaded.tab') + suffix)

    def handle_received_ids(
            self,
            ids: List[str],
            ext: Optional[str] = None
    ) -> list:
        """
        Handle (download) received message files

        :param ids: list of file IDs
        :param ext: file extension to add (optional)
        :return: downloaded files paths
        """
        num_downloaded = 0
        paths = []
        for file_id in ids:
            self.log("Downloading file: {}".format(file_id))
            path = self.window.controller.attachment.download(file_id, ext=ext)
            if path is not None:
                paths.append(path)
                num_downloaded += 1

        if num_downloaded > 0:
            pass
            # show alert with downloaded files
            # msg = "Downloaded {} file(s): {}".format(num_downloaded, ", ".join(paths))
            # self.window.ui.dialogs.alert(msg)

        return paths

    def is_log(self) -> bool:
        """
        Check if logging is enabled

        :return: bool
        """
        if self.window.core.config.has('log.assistants') \
                and self.window.core.config.get('log.assistants'):
            return True
        return False

    def log(self, msg: str):
        """
        Log message

        :param msg: message
        """
        if self.is_log():
            self.window.core.debug.info(msg, True)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.01.16 01:00:00                  #
# ================================================== #

from PySide6.QtGui import QAction, QIcon

from pygpt_net.utils import trans
import pygpt_net.icons_rc


class Donate:
    def __init__(self, window=None):
        """
        Menu setup

        :param window: Window instance
        """
        self.window = window

    def setup(self):
        """Setup donate menu"""
        self.window.ui.menu['donate.coffee'] = QAction(QIcon(":/icons/favorite.svg"), "Buy me a coffee",
                                                    self.window)
        self.window.ui.menu['donate.coffee'].setMenuRole(QAction.MenuRole.NoRole)
        self.window.ui.menu['donate.paypal'] = QAction(QIcon(":/icons/favorite.svg"), "PayPal",
                                                        self.window)
        self.window.ui.menu['donate.github'] = QAction(QIcon(":/icons/favorite.svg"), "GitHub Sponsors", self.window)

        self.window.ui.menu['donate.coffee'].triggered.connect(
            lambda: self.window.controller.dialogs.info.donate('coffee'))
        self.window.ui.menu['donate.paypal'].triggered.connect(
            lambda: self.window.controller.dialogs.info.donate('paypal'))
        self.window.ui.menu['donate.github'].triggered.connect(
            lambda: self.window.controller.dialogs.info.donate('github'))

        self.window.ui.menu['menu.donate'] = self.window.menuBar().addMenu(trans("menu.info.donate"))
        self.window.ui.menu['menu.donate'].addAction(self.window.ui.menu['donate.coffee'])
        self.window.ui.menu['menu.donate'].addAction(self.window.ui.menu['donate.paypal'])
        self.window.ui.menu['menu.donate'].addAction(self.window.ui.menu['donate.github'])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.11.17 03:00:00                  #
# ================================================== #

from .legacy import Legacy
from .memory import Memory
from .observer import Observer
from .provider import Provider
from .runner import Runner
from .tools import Tools

class Agents:
    def __init__(self, window=None):
        """
        Agents core

        :param window: Window instance
        """
        self.window = window
        self.legacy = Legacy(window)
        self.memory = Memory(window)
        self.observer = Observer(window)
        self.provider = Provider(window)
        self.runner = Runner(window)
        self.tools = Tools(window)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.16 01:00:00                  #
# ================================================== #

from typing import List, Dict


class Common:
    def __init__(self, window=None):
        """
        Idx common controller

        :param window: Window instance
        """
        self.window = window

    def setup(self):
        """Set up UI"""
        pass

    def get_loaders_choices(self) -> List[Dict[str, str]]:
        """
        Get available loaders choices

        Prefixes:
            file_ - file loader
            web_ - web loader
            online_ - online loader

        :return: list of available loaders
        """
        data = []

        # offline (built-in)
        loaders = self.window.core.idx.indexing.get_data_providers()
        for id in loaders:
            loader = loaders[id]
            # file
            if "file" in loader.type:
                id = "file_" + id
                ext_str = ""
                if loader.extensions:
                    ext_str = " (" + ", ".join(loader.extensions) + ")"
                # name = "(File) " + loader.name + ext_str  # TODO: implement option names
                data.append({
                    id: loader.name
                })
                # sort by name
                data = sorted(data, key=lambda x: list(x.values())[0])
            # web
            if "web" in loader.type:
                id = "web_" + id
                # name = "(Web) " + loader.name  # TODO: implement option names
                data.append({
                    id: loader.name
                })
                # sort by name
                data = sorted(data, key=lambda x: list(x.values())[0])
        return data



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.01.16 17:00:00                  #
# ================================================== #

from PySide6.QtWidgets import QDialog, QLabel, QVBoxLayout, QPushButton

from pygpt_net.ui.widget.element.labels import CmdLabel


class SnapDialogCamera(QDialog):
    def __init__(self, window=None):
        """
        Snap dialog for camera

        :param window: main window
        """
        super(SnapDialogCamera, self).__init__(window)
        self.window = window
        self.setParent(window)
        self.setWindowTitle("Snap version detected")
        self.cmd = CmdLabel(self.window, "sudo snap connect pygpt:camera")

        self.btn = QPushButton("OK")
        self.btn.clicked.connect(self.accept)
        self.btn.setStyleSheet("margin: 10px 0px 0px 0px;")

        # layout
        self.layout = QVBoxLayout()
        self.message = QLabel(
            "Camera not connected? It must be connected in the Snap environment.\n"
            "Run the following command to enable the camera and restart the application:")
        self.message.setStyleSheet("margin: 10px 0px 10px 0px;")
        self.layout.addWidget(self.message)
        self.layout.addWidget(self.cmd)
        self.layout.addWidget(self.btn)
        self.layout.addStretch()
        self.setLayout(self.layout)


class SnapDialogAudioInput(QDialog):
    def __init__(self, window=None):
        """
        Snap dialog for audio input

        :param window: main window
        """
        super(SnapDialogAudioInput, self).__init__(window)
        self.window = window
        self.setParent(window)
        self.setWindowTitle("Snap version is detected")
        self.cmd = CmdLabel(self.window, "sudo snap connect pygpt:alsa && sudo snap connect pygpt:audio-record :audio-record")

        self.btn = QPushButton("OK")
        self.btn.clicked.connect(self.accept)
        self.btn.setStyleSheet("margin: 10px 0px 0px 0px;")

        # layout
        self.layout = QVBoxLayout()
        self.message = QLabel(
            "Tip: Microphone must be manually connected in the Snap environment.\n"
            "If it is connected, click on the OK button, and this warning will not be displayed again.\n"
            "If it is NOT connected yet, run the following command and restart the application:")
        self.message.setStyleSheet("margin: 10px 0px 10px 0px;")
        self.layout.addWidget(self.message)
        self.layout.addWidget(self.cmd)
        self.layout.addWidget(self.btn)
        self.layout.addStretch()
        self.setLayout(self.layout)


class SnapDialogAudioOutput(QDialog):
    def __init__(self, window=None):
        """
        Snap dialog for audio output

        :param window: main window
        """
        super(SnapDialogAudioOutput, self).__init__(window)
        self.window = window
        self.setParent(window)
        self.setWindowTitle("Snap version is detected")
        self.cmd = CmdLabel(self.window, "sudo snap connect pygpt:alsa && sudo snap connect pygpt:audio-playback")

        self.btn = QPushButton("OK")
        self.btn.clicked.connect(self.accept)
        self.btn.setStyleSheet("margin: 10px 0px 0px 0px;")

        # layout
        self.layout = QVBoxLayout()
        self.message = QLabel(
            "Audio Device not connected? It must be connected in the Snap environment.\n"
            "Run the following command to enable the audio output and restart the application:")
        self.message.setStyleSheet("margin: 10px 0px 10px 0px;")
        self.layout.addWidget(self.message)
        self.layout.addWidget(self.cmd)
        self.layout.addWidget(self.btn)
        self.layout.addStretch()
        self.setLayout(self.layout)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.14 00:00:00                  #
# ================================================== #

import os
import configparser
import io
from typing import Optional

from pygpt_net.config import Config


class Locale:
    def __init__(
            self,
            domain: Optional[str] = None,
            config: Optional[Config] = None
    ):
        """
        Locale loader

        :param domain: translation domain
        :param config: Config instance
        """
        if config is None:
            self.config = Config()
        else:
            self.config = config
        self.lang = 'en'  # default language = en
        self.fallback = 'en'  # fallback language = en
        self.default_domain = 'locale'  # default translation domain
        self.ini_key = 'LOCALE'  # ini key for translations
        self.data = {}
        self.config.init(False)  # load config
        if self.config.has('lang'):
            self.lang = self.config.get_lang()
        self.load(self.lang, domain)

    def reload_config(self):
        """Reload configuration"""
        workdir = self.config.prepare_workdir()
        self.config.set_workdir(workdir)
        self.config.load(False)
        if self.config.has('lang'):
            self.lang = self.config.get_lang()
        self.load(self.lang)

    def reload(self, domain: Optional[str] = None):
        """
        Reload translations for domain

        :param domain: translation domain
        """
        self.config.load(False)
        if self.config.has('lang'):
            self.lang = self.config.get_lang()
        self.load(self.lang, domain)

    def from_file(self, path: str) -> dict:
        """
        Load and parse translations from file

        :param path: path to ini file
        :return: dict with translations
        """
        ini = configparser.ConfigParser()
        data = io.open(path, mode="r", encoding="utf-8")
        ini.read_string(data.read())
        return dict(ini.items(self.ini_key))

    def load_by_lang(
            self,
            lang: str,
            domain: Optional[str] = None
    ):
        """
        Load translation data by language code

        :param lang: language code
        :param domain: translation domain
        """
        id = self.default_domain
        if domain is not None:
            id = domain
        if id not in self.data:
            self.data[id] = {}
        try:
            path = self.get_base_path(id, lang)
            if os.path.exists(path):
                data = self.from_file(path)
                for key in data:
                    self.data[id][key] = data[key]
            path = self.get_user_path(id, lang)
            if os.path.exists(path):
                data = self.from_file(path)
                for key in data:
                    self.data[id][key] = data[key]
        except Exception as e:
            print(e)

    def load(
            self,
            lang: str,
            domain: Optional[str] = None
    ):
        """
        Load translation data

        :param lang: language code
        :param domain: translation domain
        """
        if type(lang) is not str:
            lang = self.fallback
        if lang != self.fallback:
            self.load_by_lang(self.fallback, domain)  # load fallback first
        self.load_by_lang(lang, domain)

    def get(
            self,
            key: str,
            domain: Optional[str] = None,
            params: Optional[dict] = None
    ) -> str:
        """
        Return translation for key and domain

        :param key: translation key
        :param domain: translation domain
        :param params: translation params dict for replacement
        :return: translated string or key if not found
        """
        id = self.default_domain
        if domain is not None:
            id = domain
        if id not in self.data:
            self.load(self.lang, domain)
        if id in self.data and key in self.data[id]:
            if type(params) is dict and len(params) > 0:
                try:
                    return self.data[id][key].replace('\\n', "\n").format(**params)
                except KeyError:
                    pass
            return self.data[id][key].replace('\\n', "\n")
        else:
            return key

    def get_base_path(
            self,
            domain: str,
            lang: str
    ) -> str:
        """
        Get base path for locale file

        :param domain: translation domain
        :param lang: language code
        :return: path to translations file
        """
        return os.path.join(self.config.get_app_path(), 'data', 'locale', domain + '.' + lang + '.ini')

    def get_user_path(
            self,
            domain: str,
            lang: str
    ) -> str:
        """
        Get user path for locale file (overwrites base path)

        :param domain: translation domain
        :param lang: language code
        :return: path to translations file
        """
        return os.path.join(self.config.get_user_path(), 'locale', domain + '.' + lang + '.ini')


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.14 22:00:00                  #
# ================================================== #

from typing import Dict, Any

from llama_index.core.agent import ReActAgent as Agent
from llama_index.core.agent.react_multimodal.step import (
    MultimodalReActAgentWorker,
)

from pygpt_net.core.types import (
    MODE_VISION,
)

from .base import BaseAgent

class ReactAgent(BaseAgent):
    def __init__(self, *args, **kwargs):
        super(ReactAgent, self).__init__(*args, **kwargs)
        self.id = "react"
        self.mode = "step"  # step|plan

    def get_agent(self, window, kwargs: Dict[str, Any]):
        """
        Return Agent provider instance

        :param window: window instance
        :param kwargs: keyword arguments
        :return: Agent provider instance
        """
        tools = kwargs.get("tools", [])
        verbose = kwargs.get("verbose", False)
        llm = kwargs.get("llm", None)
        chat_history = kwargs.get("chat_history", [])
        max_iterations = kwargs.get("max_iterations", 10)

        """
        # TODO: multimodal support
        model = kwargs.get("model", None)
        modes = model.mode if model is not None else None
        is_multimodal = False
        if MODE_VISION in modes:
            is_multimodal = True
        if is_multimodal:
            step_engine = MultimodalReActAgentWorker.from_tools(
                tools=tools,
                multi_modal_llm=llm,
                chat_history=chat_history,
                max_iterations=max_iterations,
                verbose=True,
            )
            return step_engine.as_agent()
        """
        return Agent.from_tools(
            tools=tools,
            llm=llm,
            chat_history=chat_history,
            max_iterations=max_iterations,
            verbose=verbose,
        )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.11.21 22:00:00                  #
# ================================================== #

import os.path
import re
import subprocess
import docker

from pygpt_net.item.ctx import CtxItem


class Runner:
    def __init__(self, plugin=None):
        """
        Cmd Runner

        :param plugin: plugin
        """
        self.plugin = plugin
        self.signals = None

    def attach_signals(self, signals):
        """
        Attach signals

        :param signals: signals
        """
        self.signals = signals

    def handle_result(self, stdout, stderr):
        """
        Handle result from subprocess

        :param stdout: stdout
        :param stderr: stderr
        :return: result
        """
        result = None
        if stdout:
            result = stdout.decode("utf-8")
            self.log("STDOUT: {}".format(result))
        if stderr:
            result = stderr.decode("utf-8")
            self.log("STDERR: {}".format(result))
        if result is None:
            result = "No result (STDOUT/STDERR empty)"
            self.log(result)
        return result

    def handle_result_docker(self, response) -> str:
        """
        Handle result from docker container

        :param response: response
        :return: result
        """
        result = None
        if response:
            result = response.decode('utf-8')
        self.log(
            "Result: {}".format(result),
            sandbox=True,
        )
        return result

    def is_sandbox(self) -> bool:
        """
        Check if sandbox is enabled

        :return: True if sandbox is enabled
        """
        return self.plugin.get_option_value('sandbox_docker')

    def get_docker(self) -> docker.client.DockerClient:
        """
        Get docker client

        :return: docker client instance
        """
        return docker.from_env()

    def get_volumes(self) -> dict:
        """
        Get docker volumes

        :return: docker volumes
        """
        path = self.plugin.window.core.config.get_user_dir('data')
        mapping = {}
        mapping[path] = {
            "bind": "/data",
            "mode": "rw",
        }
        return mapping

    def run_docker(self, cmd: str) -> bytes or None:
        """
        Run docker container with command and return response

        :param cmd: command to run
        :return: response
        """
        client = self.get_docker()
        mapping = self.get_volumes()
        try:
            response = self.plugin.docker.execute(cmd)
        except Exception as e:
            # self.error(e)
            response = str(e).encode("utf-8")
        return response

    def sys_exec_host(self, ctx: CtxItem, item: dict, request: dict) -> dict:
        """
        Execute system command on host

        :param ctx: CtxItem
        :param item: command item
        :param request: request item
        :return: response dict
        """
        msg = "Executing system command: {}".format(item["params"]['command'])
        self.log(msg)
        self.log("Running command: {}".format(item["params"]['command']))
        try:
            process = subprocess.Popen(
                item["params"]['command'],
                shell=True,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )
            stdout, stderr = process.communicate()
        except Exception as e:
            self.error(e)
            stdout = None
            stderr = str(e).encode("utf-8")
        result = self.handle_result(stdout, stderr)
        return {
            "request": request,
            "result": str(result),
            "context": "SYS OUTPUT:\n--------------------------------\n" + self.parse_result(result),
        }

    def sys_exec_sandbox(self, ctx: CtxItem, item: dict, request: dict) -> dict:
        """
        Execute system command in sandbox (docker)

        :param ctx: CtxItem
        :param item: command item
        :param request: request item
        :return: response dict
        """
        msg = "Executing system command: {}".format(item["params"]['command'])
        self.log(msg, sandbox=True)
        self.log(
            "Running command: {}".format(item["params"]['command']),
            sandbox=True,
        )
        response = self.run_docker(item["params"]['command'])
        result = self.handle_result_docker(response)
        return {
            "request": request,
            "result": str(result),
            "context": "SYS OUTPUT:\n--------------------------------\n" + self.parse_result(result),
        }

    def parse_result(self, result):
        """
        Parse result

        :param result: result
        :return: parsed result
        """
        if result is None:
            return ""
        img_ext = ["png", "jpg", "jpeg", "gif", "bmp", "tiff"]
        if result.strip().split(".")[-1].lower() in img_ext:
            path = self.prepare_path(result.strip().replace("file://", ""), on_host=True)
            if os.path.isfile(path):
                return "![Image](file://{})".format(path)
        return str(result)

    def is_absolute_path(self, path: str) -> bool:
        """
        Check if path is absolute

        :param path: path to check
        :return: True if absolute
        """
        return os.path.isabs(path)

    def prepare_path(self, path: str, on_host: bool = True) -> str:
        """
        Prepare path

        :param path: path to prepare
        :param on_host: is on host
        :return: prepared path
        """
        if self.is_absolute_path(path):
            return path
        else:
            if not self.is_sandbox() or on_host:
                return os.path.join(
                    self.plugin.window.core.config.get_user_dir('data'),
                    path,
                )
            else:
                return path

    def error(self, err: any):
        """
        Log error message

        :param err: exception or error message
        """
        if self.signals is not None:
            self.signals.error.emit(err)

    def status(self, msg: str):
        """
        Send status message

        :param msg: status message
        """
        if self.signals is not None:
            self.signals.status.emit(msg)

    def debug(self, msg: any):
        """
        Log debug message

        :param msg: message to log
        """
        if self.signals is not None:
            self.signals.debug.emit(msg)

    def log(self, msg, sandbox: bool = False):
        """
        Log message to console

        :param msg: message to log
        :param sandbox: is sandbox mode
        """
        prefix = ''
        if sandbox:
            prefix += '[DOCKER]'
        full_msg = prefix + ' ' + str(msg)

        if self.signals is not None:
            self.signals.log.emit(full_msg)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.11.05 23:00:00                  #
# ================================================== #

import os


class SyntaxHighlight:
    def __init__(self, window=None):
        self.window = window

    def exists(self, style: str) -> bool:
        """
        Check if style exists
        :param style: Style name
        :return: True if exists
        """
        return os.path.exists(os.path.join(self.window.core.config.get_app_path(), "data", "js", "highlight", "styles",
                                           f"{style}.min.css"))

    def get_style(self) -> str:
        """
        Get current style
        :return: Style name
        """
        current = self.window.core.config.get("render.code_syntax")
        if not self.exists(current):
            current = "default"
        return current

    def get_style_defs(self) -> str:
        """
        Get style definitions
        :return: Style definitions
        """
        style = self.get_style()
        path = os.path.join(self.window.core.config.get_app_path(), "data", "js", "highlight", "styles",
                            f"{style}.min.css")
        with open(path, "r") as f:
            return f.read()

    def get_styles(self) -> list:
        """
        Get available styles
        :return: Styles list
        """
        dir = os.path.join(self.window.core.config.get_app_path(), "data", "js", "highlight", "styles")
        styles = [f.replace(".min.css", "").replace(".css", "") for f in os.listdir(dir) if f.endswith(".css")]
        styles = list(dict.fromkeys(styles))
        styles.sort()
        return styles



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This store is a part of PYGPT package              #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.14 22:00:00                  #
# ================================================== #

import time
from typing import Dict

from sqlalchemy import text

from pygpt_net.item.assistant import AssistantStoreItem

from .utils import pack_item_value, unpack_store

class Storage:
    def __init__(self, window=None):
        """
        Initialize storage instance

        :param window: Window instance
        """
        self.window = window

    def attach(self, window):
        """
        Attach window instance

        :param window: Window instance
        """
        self.window = window

    def get_all(self) -> Dict[str, AssistantStoreItem]:
        """
        Return dict with AssistantStoreItem objects, indexed by ID

        :return: dict of AssistantStoreItem objects
        """
        stmt = text("""
            SELECT * FROM remote_store
        """)
        items = {}
        db = self.window.core.db.get_db()
        with db.connect() as conn:
            result = conn.execute(stmt)
            for row in result:
                store = AssistantStoreItem()
                unpack_store(store, row._asdict())
                items[store.id] = store  # index by store ID
        return items

    def get_by_id(self, id: int) -> AssistantStoreItem:
        """
        Return AssistantStoreItem by ID

        :param id: store item ID
        :return: AssistantStoreItem
        """
        stmt = text("""
            SELECT * FROM remote_store WHERE id = :id LIMIT 1
        """).bindparams(id=id)
        store = AssistantStoreItem()
        db = self.window.core.db.get_db()
        with db.connect() as conn:
            result = conn.execute(stmt)
            for row in result:
                unpack_store(store, row._asdict())
        return store

    def get_by_store_id(self, id: str) -> AssistantStoreItem:
        """
        Return AssistantStoreItem by ID

        :param id: store_id
        :return: AssistantStoreItem
        """
        stmt = text("""
            SELECT * FROM remote_store WHERE store_id = :id LIMIT 1
        """).bindparams(id=id)
        store = AssistantStoreItem()
        db = self.window.core.db.get_db()
        with db.connect() as conn:
            result = conn.execute(stmt)
            for row in result:
                unpack_store(store, row._asdict())
        return store

    def truncate_all(self) -> bool:
        """
        Truncate all stores items

        :return: True if truncated
        """
        db = self.window.core.db.get_db()
        with db.begin() as conn:
            conn.execute(text("DELETE FROM remote_store"))
            conn.execute(text("DELETE FROM sqlite_sequence WHERE name='remote_store'"))
        return True

    def delete_by_id(self, id: int) -> bool:
        """
        Delete store item by IDx

        :param id: store item ID
        :return: True if deleted
        """
        stmt = text("""
            DELETE FROM remote_store WHERE id = :id
        """).bindparams(id=id)
        db = self.window.core.db.get_db()
        with db.begin() as conn:
            conn.execute(stmt)
            return True

    def delete_by_store_id(self, id: str) -> bool:
        """
        Delete store item by IDx

        :param id: store_id
        :return: True if deleted
        """
        stmt = text("""
            DELETE FROM remote_store WHERE store_id = :id
        """).bindparams(id=id)
        db = self.window.core.db.get_db()
        with db.begin() as conn:
            conn.execute(stmt)
            return True

    def save(self, store: AssistantStoreItem):
        """
        Insert or update store item

        :param store: AssistantStoreItem object
        """
        db = self.window.core.db.get_db()
        with db.begin() as conn:
            ts = int(time.time())
            stmt = text("""
                UPDATE remote_store
                SET
                    name = :name,
                    description = :description,
                    expire_days = :expire_days,
                    usage_bytes = :usage_bytes,
                    num_files = :num_files,
                    store_id = :store_id,
                    status = :status,
                    status_json = :status_json,
                    is_thread = :is_thread,
                    last_active_ts = :last_active_ts,
                    last_sync_ts = :last_sync_ts,
                    updated_ts = :updated_ts
                WHERE id = :id
                """).bindparams(
                id=store.record_id,
                name=store.name,
                description=store.description,
                expire_days=store.expire_days,
                usage_bytes=store.usage_bytes,
                num_files=store.num_files,
                store_id=store.id,
                status=store.last_status,
                status_json=pack_item_value(store.status),
                is_thread=store.is_thread,
                last_active_ts=store.last_active,
                last_sync_ts=store.last_sync,
                updated_ts=ts
            )
            conn.execute(stmt)

    def insert(self, store: AssistantStoreItem) -> int:
        """
        Insert store item

        :param store: AssistantStoreItem object
        :return: store item ID
        """
        db = self.window.core.db.get_db()
        ts = int(time.time())
        stmt = text("""
            INSERT INTO remote_store
            (
                store_id,
                uuid,
                name,
                description,
                expire_days,
                usage_bytes,
                num_files,
                status,
                status_json,
                is_thread,
                last_active_ts,
                last_sync_ts,
                created_ts,
                updated_ts
            )
            VALUES
            (
                :store_id,
                :uuid,
                :name,
                :description,
                :expire_days,
                :usage_bytes,
                :num_files,
                :status,
                :status_json,
                :is_thread,
                :last_active_ts,
                :last_sync_ts,
                :created_ts,
                :updated_ts
            )
            """).bindparams(
            store_id=store.id,
            uuid=store.uuid,
            name=store.name,
            description=store.description,
            expire_days=int(store.expire_days or 0),
            usage_bytes=int(store.usage_bytes or 0),
            num_files=int(store.num_files or 0),
            status=store.last_status,
            status_json=pack_item_value(store.status),
            is_thread=int(store.is_thread or 0),
            last_active_ts=int(store.last_active or 0),
            last_sync_ts=int(store.last_sync or 0),
            created_ts=int(ts or 0),
            updated_ts=int(ts or 0),
        )
        with db.begin() as conn:
            result = conn.execute(stmt)
            store.record_id = result.lastrowid
            return store.record_id



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.02.02 02:00:00                  #
# ================================================== #

import copy
import json
import re
from typing import Optional, Dict, Any, List

from pygpt_net.core.types import (
    MODE_ASSISTANT,
    MODE_COMPLETION,
    MODE_LANGCHAIN,
    MODE_LLAMA_INDEX,
    MODE_AUDIO,
)
from pygpt_net.core.events import Event
from pygpt_net.item.ctx import CtxItem
from pygpt_net.item.model import ModelItem


class Command:
    DESC_LIMIT = 1024

    def __init__(self, window=None):
        """
        Commands core

        :param window: Window instance
        """
        self.window = window

    def append_syntax(
            self,
            data: Dict[str, Any],
            mode: str = None,
            model: ModelItem = None
    ) -> str:
        """
        Append command syntax to the system prompt

        :param data: event data
        :param mode: mode
        :param model: model item
        :return: prompt with appended syntax
        """
        prompt = data['prompt']
        cmd_prompt = self.window.core.prompt.get('cmd')
        extra = ""
        schema = self.extract_syntax(data['cmd'])
        if schema:
            if self.window.core.config.get('mode') == MODE_ASSISTANT:
                extra = self.window.core.prompt.get('cmd.extra.assistants')  # Assistants API env fix
            else:
                extra = self.window.core.prompt.get('cmd.extra')
        if prompt.strip() != "":
            prompt += "\n\n"
        prompt += cmd_prompt.strip().replace("{extra}", extra).replace("{schema}", schema)
        return prompt

    def extract_syntax(
            self,
            cmds: List[Dict[str, Any]]
    ) -> str:
        """
        Extract syntax from commands

        :param cmds: commands list
        :return: JSON string with commands usage syntax
        """
        data = {}
        cmds = copy.deepcopy(cmds)  # make copy to prevent changes in original data
        self.window.core.ctx.current_cmd = copy.deepcopy(cmds)  # for debug purposes

        for cmd in cmds:
            if "cmd" in cmd and "instruction" in cmd:
                cmd_name = cmd["cmd"]
                data[cmd_name] = {
                    "help": cmd["instruction"],
                }
                if "params" in cmd and len(cmd["params"]) > 0:
                    data[cmd_name]["params"] = {}
                    for param in cmd["params"]:
                        try:
                            if isinstance(param, dict):
                                if "name" in param:
                                    # assign param
                                    key = param["name"]
                                    del param["name"]
                                    data[cmd_name]["params"][key] = param

                                    # remove required, leave only optional
                                    if "required" in data[cmd_name]["params"][key]:
                                        if data[cmd_name]["params"][key]["required"] is False:
                                            data[cmd_name]["params"][key]["optional"] = True
                                        del data[cmd_name]["params"][key]["required"]

                                    # remove type if str (default)
                                    if ("type" in data[cmd_name]["params"][key]
                                            and data[cmd_name]["params"][key]["type"] == "str"):
                                        del data[cmd_name]["params"][key]["type"]

                                    # remove description and move to help
                                    if "description" in data[cmd_name]["params"][key]:
                                        data[cmd_name]["params"][key]["help"] = data[cmd_name]["params"][key][
                                            "description"]
                                        del data[cmd_name]["params"][key]["description"]

                        except Exception as e:
                            pass

        self.window.core.ctx.current_cmd_schema = data  # for debug
        return json.dumps(data)  # pack, return JSON string without indent and formatting

    def has_cmds(self, text: str) -> bool:
        """
        Check if text has command execute syntax

        :param text: text to check
        :return: True if text has commands
        """
        if text is None:
            return False
        regex_cmd = r'~###~\s*{.*}\s*~###~'
        return bool(re.search(regex_cmd, text))

    def extract_cmds(self, text: str) -> List[Dict[str, Any]]:
        """
        Extract commands from text

        :param text: text to extract commands from
        :return: list of commands (dict)
        """
        cmds = []
        try:
            chunks = text.split('~###~')
            for chunk in chunks:
                cmd = self.extract_cmd(chunk)  # extract JSON string to dict
                if cmd is not None:
                    cmds.append(cmd)  # cmd = dict
        except Exception as e:
            # do nothing
            pass
        return cmds

    def extract_cmd(self, chunk: str) -> Optional[Dict[str, Any]]:
        """
        Extract command from text chunk (JSON string)

        :param chunk: text chunk (JSON command string)
        :return: command dict
        """
        cmd = None
        chunk = chunk.strip()
        if chunk and chunk.startswith('{') and chunk.endswith('}'):
            try:
                # syntax1: {"read_file": {"path": ["my_cars.txt"]}}
                # syntax2: {"cmd": "read_file", "params": {"path": ["my_cars.txt"]}}
                cmd = json.loads(chunk)

                # if the first key is not "cmd", then try to convert from incorrect syntax into "cmd" syntax:
                if "cmd" not in cmd:
                    if len(cmd) == 1:
                        for key in cmd:
                            if "params" in cmd[key]:
                                cmd = {
                                    "cmd": key,
                                    "params": cmd[key]["params"]
                                }
                            else:
                                cmd = {
                                    "cmd": key,
                                    "params": cmd[key]
                                }
            except json.JSONDecodeError as e:
                # do nothing
                pass
        return cmd

    def from_commands(
            self,
            cmds: List[Dict[str, Any]]
    ) -> List[Dict[str, Any]]:
        """
        Unpack commands to execution list

        :param cmds: commands list
        :return parsed commands
        """
        commands = []
        for cmd in cmds:
            if 'cmd' in cmd:
                commands.append(cmd)
        return commands

    def unpack_tool_calls(
            self,
            tool_calls: List
    ) -> List[Dict[str, Any]]:
        """
        Unpack tool calls from OpenAI response

        :param tool_calls: tool calls list
        :return: parsed tool calls list
        """
        parsed = []
        for tool_call in tool_calls:
            try:
                parsed.append(
                    {
                        "id": tool_call.id,
                        "type": "function",
                        "function": {
                            "name": tool_call.function.name,
                            "arguments": json.loads(tool_call.function.arguments)
                        }
                    }
                )
            except Exception as e:
                self.window.core.debug.log(e)
                print("Error parsing tool call: " + str(e))
        return parsed

    def unpack_tool_calls_chunks(
            self,
            ctx: CtxItem,
            tool_calls: List[Dict[str, Any]]
    ):
        """
        Handle / unpack tool calls

        :param ctx: context
        :param tool_calls: tool calls
        """
        tmp_calls = []
        for tool_call in tool_calls:
            try:
                if "function" not in tool_call:
                    continue
                if "arguments" not in tool_call["function"]:
                    continue
                tool_call["function"]["arguments"] = json.loads(
                    tool_call["function"]["arguments"]
                )
                tmp_calls.append(tool_call)
            except Exception as e:
                self.window.core.debug.log(e)
                print("Error parsing tool call JSON arguments: ", tool_call["function"]["arguments"])
        ctx.tool_calls = tmp_calls

    def unpack_tool_calls_from_llama(
            self,
            tool_calls: List
    ) -> List[Dict[str, Any]]:
        """
        Unpack tool calls from Llama-index response

        :param tool_calls: tool calls list
        :return: parsed tool calls list
        """
        parsed = []
        for tool_call in tool_calls:
            try:
                parsed.append(
                    {
                        "id": tool_call.tool_id,
                        "type": "function",
                        "function": {
                            "name": tool_call.tool_name,
                            "arguments": tool_call.tool_kwargs,
                        }
                    }
                )
            except Exception as e:
                self.window.core.debug.log(e)
                print("Error parsing tool call: " + str(e))
        return parsed

    def tool_call_to_cmd(
            self,
            tool_call: Dict[str, Any]
    ) -> Dict[str, Any]:
        """
        Convert tool call to command

        :param tool_call: tool call
        :return: command
        """
        return {
            "cmd": tool_call["function"]["name"],
            "params": tool_call["function"]["arguments"]
        }

    def tool_calls_to_cmds(
            self,
            tool_calls: List[Dict[str, Any]]
    ) -> list:
        """
        Convert tool calls to commands

        :param tool_calls: tool calls
        :return: commands
        """
        cmds = []
        for tool_call in tool_calls:
            cmds.append(self.tool_call_to_cmd(tool_call))
        return cmds

    def pack_cmds(
            self,
            cmds: List[Dict[str, Any]]
    ) -> str:
        """
        Pack commands to string

        :param cmds: commands
        :return: packed commands string
        """
        packed = ""
        for cmd in cmds:
            packed += "~###~" + json.dumps(cmd) + "~###~"
        return packed

    def append_tool_calls(self, ctx: CtxItem):
        """
        Append tool calls as CMD to context output

        :param ctx: context item
        """
        cmds = self.tool_calls_to_cmds(ctx.tool_calls)
        cmds_str = self.pack_cmds(cmds)
        if ctx.output is None:
            ctx.output = ""
        ctx.output += cmds_str

    def get_tool_calls_outputs(self, ctx: CtxItem) -> list:
        """
        Prepare and get tool calls outputs to send back to assistant

        :param ctx: context item
        :return: list of tool calls outputs
        """
        outputs = []
        idx_outputs = {}
        for tool_call in ctx.tool_calls:
            call_id = tool_call["id"]
            idx_outputs[call_id] = ""
        try:
            responses = json.loads(ctx.input)  # response is JSON string in input
            for tool_call in ctx.tool_calls:
                call_id = tool_call["id"]
                for response in responses:
                    if "request" in response:
                        if "cmd" in response["request"]:
                            func_name = response["request"]["cmd"]
                            if tool_call["function"]["name"] == func_name:
                                idx_outputs[call_id] = response["result"]

        except Exception as e:
            self.window.core.debug.log(e)

        for id in idx_outputs:
            data = {
                "tool_call_id": id,
                "output": idx_outputs[id],
            }
            outputs.append(data)

        if not isinstance(ctx.extra, dict):
            ctx.extra = {}
        ctx.extra["tool_calls_outputs"] = outputs
        return outputs

    def get_functions(
            self,
            parent_id: Optional[str] = None
    ) -> List[Dict[str, Any]]:
        """
        Get current functions list

        :param parent_id: parent ID
        :return: functions list
        """
        func = []
        func_user = self.window.controller.presets.get_current_functions()
        if self.is_native_enabled():
            func = self.as_native_functions(all=False, parent_id=parent_id)
        if func_user is None:
            func_user = []
        return func + func_user  # merge both

    def as_native_functions(
            self,
            all: bool = False,
            parent_id: Optional[str] = None
    ) -> List[Dict[str, Any]]:
        """
        Convert internal functions to native API format

        :param: all True to include all
        :param: parent_id: parent context ID
        :return: native functions list
        """
        """
        # Native API format (example):
        # https://platform.openai.com/docs/guides/function-calling
        # At this moment it must be converted to format:

        functions = [
        {
            "name": "get_delivery_date",
            "description": "Get the delivery date for a customer's order. Call this whenever you need to know 
            the delivery date, for example when a customer asks 'Where is my package'",
            "params": {
                "type": "object",
                "properties": {
                    "order_id": {
                        "type": "string",
                        "description": "The customer's order ID.",
                    },
                },
                "required": ["order_id"],
                "additionalProperties": False,
            },
        }]
        """
        func_plugins = []
        func_agent = []
        func_experts = []
        data = {
            'syntax': [],
            'cmd': [],
        }

        if self.window.core.config.get('cmd') or all:
            event = Event(Event.CMD_SYNTAX, data)
            self.window.dispatch(event)
        elif self.window.controller.plugins.is_type_enabled("cmd.inline"):
            event = Event(Event.CMD_SYNTAX_INLINE, data)
            self.window.dispatch(event)

        cmds = copy.deepcopy(data['cmd'])  # make copy to prevent changes in original plugins cmd
        func_plugins = self.cmds_to_functions(cmds)  # plugin functions
        if self.window.controller.agent.legacy.enabled():
            func_agent = self.cmds_to_functions(self.window.controller.agent.legacy.get_functions())  # agent functions
        if self.window.controller.agent.experts.enabled():
            func_experts = self.cmds_to_functions(self.window.core.experts.get_functions())  # agent functions
        return func_plugins + func_agent + func_experts

    def cmds_to_functions(
            self,
            cmds: List[Dict[str, Any]]
    ) -> List[Dict[str, Any]]:
        """
        Convert commands to functions (native API format)

        :param cmds: commands list
        :return: functions list
        """
        functions = []
        for cmd in cmds:
            if "cmd" in cmd and "instruction" in cmd:
                cmd_name = cmd["cmd"]
                desc = cmd["instruction"]
                if len(desc) > self.DESC_LIMIT:
                    desc = desc[:self.DESC_LIMIT]  # limit description to 1024 characters
                functions.append(
                    {
                        "name": cmd_name,
                        "desc": desc,
                        "params": json.dumps(self.extract_params(cmd), indent=4),
                    }
                )
        return functions

    def extract_params(
            self,
            cmd: Dict[str, Any]
    ) -> Dict[str, Any]:
        """
        Extract parameters from command (to native API JSON schema format)

        :param cmd: command dict
        :return: parameters dict
        """
        required = []
        params = {
            "type": "object",
            "properties": {},
            "required": [],
            "additionalProperties": False,
        }
        if "params" in cmd and len(cmd["params"]) > 0:
            for param in cmd["params"]:
                # add required params
                if "required" in param and param["required"]:
                    required.append(param["name"])

            # extract params and convert to JSON schema format
            for param in cmd["params"]:
                try:
                    if isinstance(param, dict):
                        if "name" in param:
                            key = param["name"]
                            params["properties"][key] = {}
                            params["properties"][key]["type"] = "string"
                            params["properties"][key]["description"] = ""

                            # add required fields
                            if "type" in param:
                                params["properties"][key]["type"] = param["type"]
                            if "description" in param:
                                desc = param["description"]
                                if len(desc) > self.DESC_LIMIT:
                                    desc = desc[:self.DESC_LIMIT]  # limit description to 1024 characters
                                params["properties"][key]["description"] = desc

                            # append enum if exists
                            if "enum" in param:
                                params["properties"][key]["description"] += ", enum: " + json.dumps(
                                    param["enum"]) + ")"
                                # get dict keys from param["enum"][key] as list:
                                values = []
                                if key in param["enum"]:
                                    if isinstance(param["enum"][key], dict):  # check for sub-dicts
                                        values = list(param["enum"][key].keys())
                                        sub_values = []
                                        for k in param["enum"][key]:
                                            if isinstance(param["enum"][key][k], dict):
                                                for v in list(param["enum"][key][k].keys()):
                                                    if v not in sub_values:
                                                        sub_values.append(v)
                                            elif isinstance(param["enum"][key][k], list):
                                                for v in param["enum"][key][k]:
                                                    if v not in sub_values:
                                                        sub_values.append(v)
                                        if len(sub_values) > 0:
                                            values = sub_values
                                    elif isinstance(param["enum"][key], list):
                                        values = param["enum"][key]

                                    if values:
                                        params["properties"][key]["enum"] = values

                            # remove defaults and append to description
                            if "default" in param:
                                params["properties"][key]["description"] += ", default: " + str(
                                    param["default"]) + ")"

                            # convert internal types to supported by JSON schema
                            if params["properties"][key]["type"] == "str":
                                params["properties"][key]["type"] = "string"
                            elif params["properties"][key]["type"] == "enum":
                                params["properties"][key]["type"] = "string"
                            elif params["properties"][key]["type"] == "text":
                                params["properties"][key]["type"] = "string"
                            elif params["properties"][key]["type"] == "int":
                                params["properties"][key]["type"] = "integer"
                            elif params["properties"][key]["type"] == "bool":
                                params["properties"][key]["type"] = "boolean"
                            elif params["properties"][key]["type"] == "dict":
                                params["properties"][key]["type"] = "object"
                            elif params["properties"][key]["type"] == "list":
                                params["properties"][key]["type"] = "array"
                                params["properties"][key]["items"] = {
                                    "$ref": "#"
                                }
                except Exception as e:
                    print(e)
                    pass
        return params

    def is_native_enabled(self) -> bool:
        """
        Check if native tool calls are enabled

        :return: True if enabled
        """
        disabled_modes = [
            MODE_LLAMA_INDEX,
            MODE_LANGCHAIN,
            MODE_COMPLETION,
        ]
        mode = self.window.core.config.get('mode')
        if mode in disabled_modes:
            return False  # disabled for specific modes
        if self.window.controller.agent.legacy.enabled() or self.window.controller.agent.experts.enabled():
            return False
        return self.window.core.config.get('func_call.native', False)  # otherwise check config

    def is_enabled(self, cmd: str) -> bool:
        """
        Check if command is enabled

        :param cmd: command
        :return: True if command is enabled
        """
        enabled_cmds = []
        data = {
            'prompt': "",
            'silent': True,
            'force': True,
            'syntax': [],
            'cmd': [],
        }
        event = Event(Event.CMD_SYNTAX, data)
        self.window.dispatch(event)
        if (event.data and "cmd" in event.data
                and isinstance(event.data["cmd"], list)):
            for item in event.data["cmd"]:
                if "cmd" in item:
                    enabled_cmds.append(item["cmd"])
        data = {
            'prompt': "",
            'silent': True,
            'force': True,
            'syntax': [],
            'cmd': [],
        }
        event = Event(Event.CMD_SYNTAX_INLINE, data)
        self.window.dispatch(event)
        if (event.data and "cmd" in event.data
            and isinstance(event.data["cmd"], list)):
            for item in event.data["cmd"]:
                if "cmd" in item:
                    enabled_cmds.append(item["cmd"])
        if cmd in enabled_cmds:
            return True
        return False

    def is_model_supports_tools(
            self,
            mode: str,
            model: ModelItem = None) -> bool:
        """
        Check if model supports tools

        :param mode: mode
        :param model: model item
        :return: True if model supports tools
        """
        return True  # TMP allowed all
        if model is None:
            return False
        disabled_models = [
            "deepseek-r1:1.5b",
            "deepseek-r1:7b",
            "llama2",
            "llama3.1",
            "codellama",
        ]
        if model.id is not None:
            for disabled_model in disabled_models:
                if (model.llama_index['provider'] == "ollama"
                        and model.id.startswith(disabled_model)):
                    return False
        return True



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.11.18 21:00:00                  #
# ================================================== #

from PySide6.QtCore import Slot, Signal

from pygpt_net.plugin.base.worker import BaseWorker, BaseSignals


class WorkerSignals(BaseSignals):
    updated = Signal()  # calendar view update signal


class Worker(BaseWorker):
    def __init__(self, *args, **kwargs):
        super(Worker, self).__init__()
        self.signals = WorkerSignals()
        self.args = args
        self.kwargs = kwargs
        self.plugin = None

    @Slot()
    def run(self):
        responses = []
        msg = None
        for item in self.cmds:
            if self.is_stopped():
                break
            response = None
            try:
                if item["cmd"] == "get_ctx_list_in_date_range":
                    response = self.cmd_get_ctx_list_in_date_range(item)

                elif item["cmd"] == "get_ctx_content_by_id":
                    response = self.cmd_get_ctx_content_by_id(item)

                elif item["cmd"] == "get_day_note":
                    response = self.cmd_get_day_note(item)

                elif item["cmd"] == "add_day_note":
                    response = self.cmd_add_day_note(item)

                elif item["cmd"] == "update_day_note":
                    response = self.cmd_update_day_note(item)

                elif item["cmd"] == "remove_day_note":
                    response = self.cmd_remove_day_note(item)

                elif item["cmd"] == "count_ctx_in_date":
                    response = self.cmd_count_ctx_in_date(item)

                if response:
                    responses.append(response)

            except Exception as e:
                msg = "Error: {}".format(e)
                responses.append(
                    self.make_response(
                        item,
                        self.throw_error(e)
                    )
                )

        # send response
        if len(responses) > 0:
            self.reply_more(responses)

        # update status
        if msg is not None:
            self.status(msg)

    def cmd_get_ctx_list_in_date_range(self, item: dict) -> dict:
        """
        Get context list in date range

        :param item: command item
        :return: response item
        """
        range = self.get_param(item, "range_query", "")
        result = self.plugin.get_list(range)
        return self.make_response(item, result)

    def cmd_get_ctx_content_by_id(self, item: dict) -> dict:
        """
        Get context content by id

        :param item: command item
        :return: response item
        """
        id = int(self.get_param(item, "id", 0))
        prompt = "Summary this conversation"
        if self.has_param(item, "summary_query"):
            prompt = self.get_param(item, "summary_query")
        result = self.plugin.get_summary(id, prompt)
        return self.make_response(item, result)

    def cmd_get_day_note(self, item: dict) -> dict:
        """
        Get day note

        :param item: command item
        :return: response item
        """
        year = int(self.get_param(item, "year", 0))
        month = int(self.get_param(item, "month", 0))
        day = int(self.get_param(item, "day", 0))
        result = self.plugin.get_day_note(year, month, day)
        return self.make_response(item, result)

    def cmd_add_day_note(self, item: dict) -> dict:
        """
        Add day note

        :param item: command item
        :return: response item
        """
        year = int(self.get_param(item, "year", 0))
        month = int(self.get_param(item, "month", 0))
        day = int(self.get_param(item, "day", 0))
        note = self.get_param(item, "note", "")
        self.signals.updated.emit()
        result = self.plugin.add_day_note(year, month, day, note)
        return self.make_response(item, result)

    def cmd_update_day_note(self, item: dict) -> dict:
        """
        Update day note

        :param item: command item
        :return: response item
        """
        year = int(self.get_param(item, "year", 0))
        month = int(self.get_param(item, "month", 0))
        day = int(self.get_param(item, "day", 0))
        note = self.get_param(item, "content", "")
        self.signals.updated.emit()
        result = self.plugin.update_day_note(year, month, day, note)
        return self.make_response(item, result)

    def cmd_remove_day_note(self, item: dict) -> dict:
        """
        Remove day note

        :param item: command item
        :return: response item
        """
        year = int(self.get_param(item, "year", 0))
        month = int(self.get_param(item, "month", 0))
        day = int(self.get_param(item, "day", 0))
        self.signals.updated.emit()
        result = self.plugin.remove_day_note(year, month, day)
        return self.make_response(item, result)

    def cmd_count_ctx_in_date(self, item: dict) -> dict:
        """
        Count context in date

        :param item: command item
        :return: response item
        """
        year = None
        month = None
        day = None
        if self.has_param(item, "year"):
            tmp_year = self.get_param(item, "year", "")
            if tmp_year != "":
                year = int(tmp_year)
        if self.has_param(item, "month"):
            tmp_month = self.get_param(item, "month", "")
            if tmp_month != "":
                month = int(tmp_month)
        if self.has_param(item, "day"):
            tmp_day = self.get_param(item, "day", "")
            if tmp_day != "":
                day = int(tmp_day)
        result = self.plugin.count_ctx_in_date(year, month, day)
        return self.make_response(item, result)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.11.26 04:00:00                  #
# ================================================== #
import json


class ContextDebug:
    def __init__(self, window=None):
        """
        Context debug

        :param window: Window instance
        """
        self.window = window
        self.id = 'context'

    def update(self):
        """Update debug window"""
        self.window.core.debug.begin(self.id)
        if self.window.core.bridge.last_context is not None:
            self.window.core.debug.add(self.id, 'bridge (last call)', str(self.window.core.bridge.last_context.to_dict()))
        else:
            self.window.core.debug.add(self.id, 'bridge (last call)', '---')
        if self.window.core.bridge.last_context_quick is not None:
            self.window.core.debug.add(self.id, 'bridge (last quick call)', str(self.window.core.bridge.last_context_quick.to_dict()))
        else:
            self.window.core.debug.add(self.id, 'bridge (last quick call)', '---')
        if self.window.controller.kernel.stack.current is not None:
            self.window.core.debug.add(self.id, 'reply (queue)', str(self.window.controller.kernel.stack.current.to_dict()))
        else:
            self.window.core.debug.add(self.id, 'reply (queue)', '---')
        self.window.core.debug.add(self.id, 'reply (locked)', str(self.window.controller.kernel.stack.locked))
        self.window.core.debug.add(self.id, 'reply (processed)', str(self.window.controller.kernel.stack.processed))
        self.window.core.debug.add(self.id, 'current (id)', str(self.window.core.ctx.get_current()))
        self.window.core.debug.add(self.id, 'len(meta)', len(self.window.core.ctx.meta))
        self.window.core.debug.add(self.id, 'len(items)', len(self.window.core.ctx.get_items()))
        self.window.core.debug.add(self.id, 'SYS PROMPT (current)', str(self.window.core.ctx.current_sys_prompt))
        self.window.core.debug.add(self.id, 'CMD (current)', str(self.window.core.ctx.current_cmd))
        self.window.core.debug.add(self.id, 'CMD schema (current)', str(self.window.core.ctx.current_cmd_schema))
        self.window.core.debug.add(self.id, 'FUNCTIONS (current)', str(self.get_functions()))
        self.window.core.debug.add(self.id, 'Attachments: last used content',
                                   str(self.window.core.attachments.context.last_used_content))
        self.window.core.debug.add(self.id, 'Attachments: last used context',
                                  str(self.window.core.attachments.context.last_used_context))

        current = None
        if self.window.core.ctx.get_current() is not None:
            if self.window.core.ctx.get_current() in self.window.core.ctx.meta:
                current = self.window.core.ctx.meta[self.window.core.ctx.get_current()]
            if current is not None:
                data = current.to_dict()
                self.window.core.debug.add(self.id, '*** (current)', str(data))

        if self.window.core.ctx.get_tmp_meta() is not None:
            self.window.core.debug.add(self.id, 'tmp meta', str(self.window.core.ctx.get_tmp_meta().to_dict()))
            
        self.window.core.debug.add(self.id, 'group_id (active)', str(self.window.controller.ctx.group_id))
        self.window.core.debug.add(self.id, 'assistant', str(self.window.core.ctx.get_assistant()))
        self.window.core.debug.add(self.id, 'mode', str(self.window.core.ctx.get_mode()))
        self.window.core.debug.add(self.id, 'model', str(self.window.core.ctx.get_model()))
        self.window.core.debug.add(self.id, 'preset', str(self.window.core.ctx.get_preset()))
        self.window.core.debug.add(self.id, 'run', str(self.window.core.ctx.get_run()))
        self.window.core.debug.add(self.id, 'status', str(self.window.core.ctx.get_status()))
        self.window.core.debug.add(self.id, 'thread', str(self.window.core.ctx.get_thread()))
        self.window.core.debug.add(self.id, 'last_mode', str(self.window.core.ctx.get_last_mode()))
        self.window.core.debug.add(self.id, 'last_model', str(self.window.core.ctx.get_last_model()))
        self.window.core.debug.add(self.id, 'search_string', str(self.window.core.ctx.get_search_string()))
        self.window.core.debug.add(self.id, 'filters', str(self.window.core.ctx.filters))
        self.window.core.debug.add(self.id, 'filters_labels', str(self.window.core.ctx.filters_labels))
        self.window.core.debug.add(self.id, 'allowed_modes', str(self.window.core.ctx.allowed_modes))

        i = 0
        self.window.core.debug.add(self.id, 'items[]', '')
        for item in self.window.core.ctx.get_items():
            data = item.to_dict()
            self.window.core.debug.add(self.id, str(item.id), str(data))
            i += 1

        self.window.core.debug.end(self.id)

    def get_functions(self) -> list:
        """
        Parse functions

        :return: List of functions
        """
        parsed = []
        functions = self.window.core.command.get_functions()
        for func in functions:
            try:
                item = {
                    "name": func['name'],
                    "desc": func['desc'],
                    "params": json.loads(func['params']),
                }
                parsed.append(item)
            except Exception as e:
                pass
        return parsed



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.01.31 19:00:00                  #
# ================================================== #

import json
from typing import Optional, Dict, Any, List

from llama_index.core.llms import ChatMessage, MessageRole
from llama_index.core.prompts import ChatPromptTemplate
from llama_index.core.memory import ChatMemoryBuffer

from pygpt_net.core.types import (
    MODE_CHAT,
    MODE_AGENT_LLAMA,
)
from pygpt_net.core.bridge.context import BridgeContext
from pygpt_net.item.model import ModelItem
from pygpt_net.item.ctx import CtxItem

from .context import Context


class Chat:
    def __init__(self, window=None, storage=None):
        """
        Chat with index core

        :param window: Window instance
        """
        self.window = window
        self.storage = storage
        self.context = Context(window)
        self.tool_calls_not_allowed_providers = [
            "ollama",
            "hugging_face_api",
            "deepseek_api",
        ]

    def call(
            self,
            context: BridgeContext,
            extra: Optional[Dict[str, Any]] = None
    ) -> bool:
        """
        Call chat, complete or query mode

        :param context: Bridge context
        :param extra: Extra arguments
        :return: True if success
        """
        model = context.model
        idx_mode = context.idx_mode  # mode

        if model is None or not isinstance(model, ModelItem):  # check if model is provided
            raise Exception("Model config not provided")

        if idx_mode == "query":  # query index only (raw mode)
            return self.raw_query(
                context=context,
                extra=extra,
            )
        elif idx_mode == "retrieval":  # retrieval mode
            return self.retrieval(
                context=context,
                extra=extra,
            )

        # if not raw, check if chat mode is available
        if MODE_CHAT in model.llama_index['mode']:
            return self.chat(
                context=context,
                extra=extra,
            )
        else:
            return self.query(
                context=context,
                extra=extra,
            )  # if not, use query mode

    def raw_query(
            self,
            context: BridgeContext,
            extra: Optional[Dict[str, Any]] = None
    ) -> bool:
        """
        Raw query mode

        :param context: Bridge context
        :param extra: Extra arguments
        :return: True if success
        """
        return self.query(
            context=context,
            extra=extra,
        )

    def query(
            self,
            context: BridgeContext,
            extra: Optional[Dict[str, Any]] = None
    ) -> bool:
        """
        Query index mode (no chat, only single query) and append results to context

        :param context: Bridge context
        :param extra: Extra arguments
        :return: True if success
        """
        idx = context.idx
        model = context.model
        system_prompt = context.system_prompt_raw  # get raw system prompt, without plugin addons
        stream = context.stream
        ctx = context.ctx
        query = ctx.input  # user input
        verbose = self.window.core.config.get("log.llama", False)

        if model is None or not isinstance(model, ModelItem):
            raise Exception("Model config not provided")

        self.log("Query index...")
        self.log("Idx: {}, query: {}, model: {}".format(
            idx,
            query,
            model.id,
        ))

        index, llm = self.get_index(idx, model)
        input_tokens = self.window.core.tokens.from_llama_messages(
            query,
            [],
            model.id,
        )
        # query index
        tpl = self.get_custom_prompt(system_prompt)
        if tpl is not None:
            self.log("Query index with custom prompt: {}...".format(system_prompt))
            response = index.as_query_engine(
                llm=llm,
                streaming=stream,
                text_qa_template=tpl,
                verbose=verbose,
            ).query(query)  # query with custom sys prompt
        else:
            response = index.as_query_engine(
                llm=llm,
                streaming=stream,
                verbose=verbose,
            ).query(query)  # query with default prompt

        if response:
            if stream:
                ctx.add_doc_meta(self.get_metadata(response.source_nodes))  # store metadata
                ctx.stream = response.response_gen
                ctx.input_tokens = input_tokens
                ctx.set_output("", "")
            else:
                ctx.add_doc_meta(self.get_metadata(response.source_nodes))  # store metadata
                ctx.input_tokens = input_tokens
                ctx.output_tokens = self.window.core.tokens.from_llama_messages(
                    response.response,
                    [],
                    model.id,
                )  # calc from response
                ctx.set_output(str(response.response), "")
            return True
        return False

    def retrieval(
            self,
            context: BridgeContext,
            extra: Optional[Dict[str, Any]] = None
    ) -> bool:
        """
        Retrieve documents from index only

        :param context: Bridge context
        :param extra: Extra arguments
        :return: True if success
        """
        idx = context.idx
        model = context.model
        ctx = context.ctx
        query = ctx.input  # user input
        verbose = self.window.core.config.get("log.llama", False)

        self.log("Retrieval...")
        self.log("Idx: {}, retrieve only: {}".format(
            idx,
            query,
        ))

        index, llm = self.get_index(idx, model)
        retriever = index.as_retriever()
        nodes = retriever.retrieve(query)
        outputs = []
        self.log("Retrieved {} nodes...".format(len(nodes)))
        for node in nodes:
            outputs.append({
                "text": node.text,
                "score": node.score,
            })
        if outputs:
            response = ""
            for output in outputs:
                response += "**Score: {}**\n\n{}".format(output["score"], output["text"])
                if output != outputs[-1]:
                    response += "\n\n-------\n\n"
            ctx.set_output(response)
            ctx.add_doc_meta(self.get_metadata(nodes))
        return True

    def chat(
            self,
            context: BridgeContext,
            extra: Optional[Dict[str, Any]] = None
    ) -> bool:
        """
        Chat mode (conversation, using context from index) and append result to the context

        :param context: Bridge context
        :param extra: Extra arguments
        """
        idx = context.idx
        model = context.model
        system_prompt = context.system_prompt  # get final system prompt
        stream = context.stream
        ctx = context.ctx
        query = ctx.input  # user input
        chat_mode = self.window.core.config.get("llama.idx.chat.mode")
        use_index = True
        verbose = self.window.core.config.get("log.llama", False)
        allow_native_tool_calls = True
        if ('provider' in model.llama_index
                and model.llama_index['provider'] in self.tool_calls_not_allowed_providers):
            allow_native_tool_calls = False

        if idx is None or idx == "_":
            chat_mode = "simple"  # do not use query engine if no index
            use_index = False

        # disable index if no api key
        if self.window.core.config.get("api_key") == "" and self.window.core.config.get("llama.idx.embeddings.provider") == "openai":
            print("Warning: no api key! Disabling index...")
            chat_mode = "simple"  # do not use query engine if no index
            use_index = False

        if model is None or not isinstance(model, ModelItem):
            raise Exception("Model config not provided")

        self.log("Chat with index...")
        self.log("Idx: {}, query: {}, chat_mode: {}, model: {}".format(
            idx,
            query,
            chat_mode,
            model.id,
        ))

        # use index only if idx is not empty, otherwise use only LLM
        index = None
        if use_index:
            index, llm = self.get_index(idx, model)
        else:
            llm = self.window.core.idx.llm.get(model)

        # check if index is empty
        if index:
            nodes = index.docstore.docs.values()
            if not nodes:
                use_index = False

        # TODO: if multimodal support, try to get multimodal provider
        # if model.is_multimodal():
            # llm = self.window.core.idx.llm.get(model, multimodal=True)  # get multimodal LLM model

        # append context from DB
        history = self.context.get_messages(
            ctx.input,
            system_prompt,
            context.history,
        )
        memory = self.get_memory_buffer(history, llm)
        input_tokens = self.window.core.tokens.from_llama_messages(
            query,
            history,
            model.id,
        )

        if use_index:
            # TOOLS: commands are applied to system prompt here
            # index as query engine
            chat_engine = index.as_chat_engine(
                llm=llm,
                chat_mode=chat_mode,
                memory=memory,
                verbose=verbose,
                system_prompt=system_prompt,
            )
            if stream:
                response = chat_engine.stream_chat(query)
            else:
                response = chat_engine.chat(query)
        else:
            # TOOLS: commands are applied to system prompt here
            # prepare tools (native calls if enabled)
            tools = self.window.core.agents.tools.prepare(context, extra)

            # without index, use LLM as chat engine
            history.insert(0, self.context.add_system(system_prompt))
            history.append(self.context.add_user(query))
            if stream:
                # IMPORTANT: stream chat with tools not supported by all providers
                if allow_native_tool_calls and hasattr(llm, "stream_chat_with_tools"):
                    response = llm.stream_chat_with_tools(
                        tools=tools,
                        messages=history,
                    )
                else:
                    response = llm.stream_chat(
                        messages=history,
                    )
            else:
                # IMPORTANT: stream chat with tools not supported by all providers
                if allow_native_tool_calls and hasattr(llm, "chat_with_tools"):
                    response = llm.chat_with_tools(
                        tools=tools,
                        messages=history,
                    )
                else:
                    response = llm.chat(
                        messages=history,
                    )

        # handle response
        if response:
            if use_index:
                # from index as query engine
                if stream:
                    ctx.stream = response.response_gen
                    ctx.input_tokens = input_tokens
                    ctx.set_output("", "")
                    ctx.add_doc_meta(self.get_metadata(response.source_nodes))  # store metadata
                else:
                    ctx.input_tokens = input_tokens
                    ctx.output_tokens = self.window.core.tokens.from_llama_messages(
                        response.response,
                        [],
                        model.id,
                    )  # calc from response
                    output = response.response
                    if output is None:
                        output = ""
                    ctx.set_output(output, "")
                    ctx.add_doc_meta(self.get_metadata(response.source_nodes))  # store metadata
            else:
                # from LLM directly, no index
                if stream:
                    # tools are handled in stream output controller
                    ctx.stream = response  # chunk is in response.delta
                    ctx.input_tokens = input_tokens
                    ctx.set_output("", "")
                else:
                    # unpack tool calls
                    tool_calls = llm.get_tool_calls_from_response(
                        response,
                        error_on_no_tool_call=False,
                    )
                    ctx.tool_calls = self.window.core.command.unpack_tool_calls_from_llama(tool_calls)
                    output = response.message.content
                    if output is None:
                        output = ""
                    ctx.set_output(output, "")
                    ctx.input_tokens = input_tokens
                    ctx.output_tokens = self.window.core.tokens.from_llama_messages(
                        output,
                        [],
                        model.id,
                    ) # calc from response
            return True
        return False

    def query_file(
            self,
            ctx: CtxItem,
            path: str,
            query: str,
            model: Optional[ModelItem] = None
    ) -> str:
        """
        Query file using temp index (created on the fly)

        :param ctx: context
        :param path: path to file to index (in memory)
        :param query: query
        :param model: model
        :return: response
        """
        if model is None:
            model = self.window.core.models.from_defaults()

        llm, embed_model = self.window.core.idx.llm.get_service_context(model=model)
        tmp_id, index = self.storage.get_tmp(path, llm, embed_model)  # get or create tmp index

        idx = "tmp:{}".format(path)  # tmp index id
        self.log("Indexing to temporary in-memory index: {}...".format(idx))

        # index file to tmp index
        files, errors = self.window.core.idx.indexing.index_files(
            idx=idx,
            index=index,
            path=path,
            is_tmp=True,  # do not try to remove old doc id
        )

        # query tmp index
        output = None
        if len(files) > 0:
            self.log("Querying temporary in-memory index: {}...".format(idx))
            response = index.as_query_engine(
                llm=llm,
                streaming=False,
            ).query(query)  # query with default prompt
            if response:
                ctx.add_doc_meta(self.get_metadata(response.source_nodes))  # store metadata
                output = response.response

        # clean tmp index
        self.log("Removing temporary in-memory index: {} ({})...".format(idx, tmp_id))
        self.storage.clean_tmp(tmp_id)  # clean memory
        self.log("Returning response: {}".format(output))
        return output

    def query_web(
            self,
            ctx: CtxItem,
            type: str,
            url: str,
            args: Dict[str, Any],
            query: str,
            model: Optional[ModelItem] = None
    ) -> str:
        """
        Query web using temp index (created on the fly)

        :param ctx: context
        :param type: type of content
        :param url: url to index (in memory)
        :param args: extra args
        :param query: query
        :param model: model
        :return: response
        """
        parts = {
            "type": type,
            "url": url,
            "args": args,
        }
        id = json.dumps(parts)
        if model is None:
            model = self.window.core.models.from_defaults()
        llm, embed_model = self.window.core.idx.llm.get_service_context(model=model)
        tmp_id, index = self.storage.get_tmp(id, llm, embed_model)  # get or create tmp index

        idx = "tmp:{}".format(id)  # tmp index id
        self.log("Indexing to temporary in-memory index: {}...".format(idx))

        # index file to tmp index
        num, errors = self.window.core.idx.indexing.index_urls(
            idx=id,
            index=index,
            urls=[url],
            type=type,
            extra_args=args,
            is_tmp=True,  # do not try to remove old doc id
        )

        # query tmp index
        output = None
        if num > 0:
            self.log("Querying temporary in-memory index: {}...".format(idx))
            response = index.as_query_engine(
                llm=llm,
                streaming=False,
            ).query(query)  # query with default prompt
            if response:
                ctx.add_doc_meta(self.get_metadata(response.source_nodes))  # store metadata
                output = response.response

        # clean tmp index
        self.log("Removing temporary in-memory index: {} ({})...".format(idx, tmp_id))
        self.storage.clean_tmp(tmp_id)  # clean memory
        self.log("Returning response: {}...".format(output))
        return output

    def query_attachment(
            self,
            query: str,
            path: str,
            model: Optional[ModelItem] = None,
            history: Optional[List[CtxItem]] = None,
            verbose: bool = False,
    ) -> str:
        """
        Query attachment

        :param query: query
        :param path: path to index
        :param model: model
        :param history: chat history
        :param verbose: verbose mode
        :return: response
        """
        if model is None:
            model = self.window.core.models.from_defaults()
        llm, embed_model = self.window.core.idx.llm.get_service_context(model=model)
        index = self.storage.get_ctx_idx(path, llm, embed_model)

        # 1. try to retrieve directly from index
        retriever = index.as_retriever()
        nodes = retriever.retrieve(query)
        response = ""
        score = 0
        for node in nodes:
            if node.score > 0.5:
                score = node.score
                response = node.text
                break
        output = ""
        if response:
            output = str(response)
            if verbose:
                print("Found using retrieval, {} (score: {})".format(output, score))
        else:
            if verbose:
                print("Not found using retrieval, trying with query engine...")
            history = self.context.get_messages(
                query,
                "",
                history,
            )
            memory = self.get_memory_buffer(history, llm)
            response = index.as_chat_engine(
                llm=llm,
                streaming=False,
                memory=memory,
            ).chat(query)
            if response:
                output = str(response.response)
        return output

    def query_retrieval(
            self,
            query: str,
            idx: str,
            model: Optional[ModelItem] = None
    ) -> str:
        """
        Query attachment

        :param query: query
        :param idx: index id
        :param model: model
        :return: response
        """
        if model is None:
            model = self.window.core.models.from_defaults()
        index, llm = self.get_index(idx, model)
        retriever = index.as_retriever()
        nodes = retriever.retrieve(query)
        response = ""
        for node in nodes:
            if node.score > 0.5:
                response = node.text
                break
        output = ""
        if response:
            output = str(response)
        return output

    def get_memory_buffer(
            self,
            history: List[ChatMessage],
            llm = None
    ) -> ChatMemoryBuffer:
        """
        Get memory buffer

        :param history: list with chat history (ChatMessage)
        :param llm: LLM model
        :return: memory buffer with chat history
        """
        return ChatMemoryBuffer.from_defaults(
            chat_history=history,
            llm=llm,
        )

    def get_custom_prompt(
            self,
            prompt: Optional[str] = None
    ) -> Optional[ChatPromptTemplate]:
        """
        Get custom prompt template if sys prompt is not empty

        :param prompt: system prompt (optional)
        :return: ChatPromptTemplate or None if prompt is empty
        """
        if prompt is None or prompt.strip() == "":
            return None

        qa_msgs = [
            ChatMessage(
                role=MessageRole.SYSTEM,
                content=prompt,
            ),
            ChatMessage(
                role=MessageRole.USER,
                content=(
                    "Context information is below.\n"
                    "---------------------\n"
                    "{context_str}\n"
                    "---------------------\n"
                    "Given the context information and not prior knowledge, "
                    "answer the question: {query_str}\n"
                ),
            ),
        ]
        return ChatPromptTemplate(qa_msgs)

    def get_index(
            self,
            idx: str,
            model: ModelItem
    ):
        """
        Get index instance

        :param idx: idx name (id)
        :param model: model instance
        :return:
        """
        # check if index exists
        if not self.storage.exists(idx):
            if idx is None:
                # create empty in memory idx
                llm, embed_model = self.window.core.idx.llm.get_service_context(model=model)
                index = self.storage.index_from_empty(embed_model)
                return index, llm
            # raise Exception("Index not prepared")

        llm, embed_model = self.window.core.idx.llm.get_service_context(model=model)
        index = self.storage.get(idx, llm, embed_model)  # get index
        return index, llm

    def get_metadata(
            self,
            source_nodes: Optional[list]
    ) -> Dict[str, Any]:
        """
        Get metadata from source nodes

        :param source_nodes: source nodes
        :return: metadata
        """
        if (source_nodes is None
                or not isinstance(source_nodes, list)
                or len(source_nodes) == 0):
            return {}
        metadata = {}
        i = 1
        max = 3
        min_score = 0.3
        for node in source_nodes:
            if hasattr(node, "id_"):
                id = node.id_
                if node.metadata is not None:
                    score = node.get_score()
                    if score > min_score:
                        metadata[id] = node.metadata
                        metadata[id]["score"] = score
                        i += 1
                        if i > max:
                            break
        return metadata

    def parse_metadata(
            self,
            metadata: Optional[Dict]
    ) -> Dict[str, Any]:
        """
        Parse metadata

        :param metadata: metadata
        :return: metadata
        """
        if (metadata is None
                or not isinstance(metadata, dict)
                or len(metadata) == 0):
            return {}
        meta = {}
        for node in metadata:
            doc_id = list(node.keys())[0]
            meta[doc_id] = node[doc_id]
            break
        return meta

    def log(self, msg: str):
        """
        Log info message

        :param msg: message
        """
        # disabled logging for thread safety
        if self.window.core.config.get("mode") == MODE_AGENT_LLAMA:
            return
        is_log = False
        if self.window.core.config.has("log.llama") \
                and self.window.core.config.get("log.llama"):
            is_log = True
        self.window.core.debug.info(msg, not is_log)
        if is_log:
            print("[LLAMA-INDEX] {}".format(msg))
        self.window.idx_logger_message.emit(msg)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.01.16 01:00:00                  #
# ================================================== #

import os.path
from typing import Optional, Union

from llama_index.core.llms.llm import BaseLLM
from llama_index.core.multi_modal_llms import MultiModalLLM
from llama_index.core.base.embeddings.base import BaseEmbedding
from llama_index.llms.openai import OpenAI

from pygpt_net.core.types import (
    MODE_LLAMA_INDEX,
)
from pygpt_net.item.model import ModelItem


class Llm:
    def __init__(self, window=None):
        """
        LLM provider core

        :param window: Window instance
        """
        self.window = window
        self.default_model = "gpt-3.5-turbo"
        self.default_embed = "openai"
        self.initialized = False

    def init(self):
        """Init base ENV vars"""
        os.environ['OPENAI_API_KEY'] = str(self.window.core.config.get('api_key'))
        os.environ['OPENAI_API_BASE'] = str(self.window.core.config.get('api_endpoint'))
        os.environ['OPENAI_ORGANIZATION'] = str(self.window.core.config.get('organization_key'))

    def get(
            self,
            model: Optional[ModelItem] = None,
            multimodal: bool = False
    ) -> Union[BaseLLM, MultiModalLLM]:
        """
        Get LLM provider

        :param model: Model item
        :param multimodal: Allow multi-modal flag (True to get multimodal provider if available)
        :return: Llama LLM instance
        """
        # TMP: deprecation warning fix
        # https://github.com/DataDog/dd-trace-py/issues/8212#issuecomment-1971063988
        if not self.initialized:
            import warnings
            from langchain._api import LangChainDeprecationWarning
            warnings.simplefilter("ignore", category=LangChainDeprecationWarning)
            self.initialized = True

        llm = None
        if model is not None:
            if 'provider' in model.llama_index:
                provider = model.llama_index['provider']
                if provider in self.window.core.llm.llms:
                    # init env vars
                    self.window.core.llm.llms[provider].init(
                        window=self.window,
                        model=model,
                        mode=MODE_LLAMA_INDEX,
                        sub_mode="",
                    )
                    # get llama LLM instance
                    if multimodal and model.is_multimodal():
                        # at first, try to get multimodal provider
                        llm = self.window.core.llm.llms[provider].llama_multimodal(
                            window=self.window,
                            model=model,
                        )
                        if llm is not None:
                            print("Using multimodal.")

                    if llm is None:
                        # if no multimodal, get default llama provider
                        llm = self.window.core.llm.llms[provider].llama(
                            window=self.window,
                            model=model,
                        )

        # default model
        if llm is None:
            self.init()  # init env vars
            llm = OpenAI(
                temperature=0.0,
                model=self.default_model,
            )
        return llm

    def get_embeddings_provider(self) -> BaseEmbedding:
        """
        Get current embeddings provider

        :return: Llama embeddings provider instance
        """
        provider = self.window.core.config.get("llama.idx.embeddings.provider", self.default_embed)
        env = self.window.core.config.get("llama.idx.embeddings.env", [])
        args = self.window.core.config.get("llama.idx.embeddings.args", [])

        if provider is None or provider not in self.window.core.llm.llms:
            provider = self.default_embed

        self.window.core.llm.llms[provider].init_embeddings(
            window=self.window,
            env=env,
        )
        return self.window.core.llm.llms[provider].get_embeddings_model(
            window=self.window,
            config=args,
        )

    def get_service_context(
            self,
            model: Optional[ModelItem] = None
    ):
        """
        Get service context + embeddings provider

        :param model: Model item (for query)
        :return: Service context instance
        """
        llm = self.get(model=model)
        embed_model = self.get_embeddings_provider()
        return llm, embed_model



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.11.18 00:00:00                  #
# ================================================== #

from .analyzer import Analyzer

class Vision:
    def __init__(self, window=None):
        """
        Audio analyzer

        :param window: Window instance
        """
        self.window = window
        self.analyzer = Analyzer(window)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.11.25 02:00:00                  #
# ================================================== #

from pygpt_net.plugin.base.config import BaseConfig, BasePlugin


class Config(BaseConfig):
    def __init__(self, plugin: BasePlugin = None, *args, **kwargs):
        super(Config, self).__init__(plugin)
        self.plugin = plugin

    def from_defaults(self, plugin: BasePlugin = None):
        """
        Set default options for plugin

        :param plugin: plugin instance
        """
        dockerfile = '''
        # Tip: After making changes to this Dockerfile, you must rebuild the image to apply the changes(Menu -> Tools -> Rebuild IPython Docker Image)

        FROM python:3.9

        # You can customize the packages installed by default here:
        # ========================================================
        RUN pip install jupyter ipykernel
        # ========================================================

        RUN mkdir /data

        # Expose the necessary ports for Jupyter kernel communication
        EXPOSE 5555 5556 5557 5558 5559

        # Data directory, bound as a volume to the local 'data/ipython' directory
        WORKDIR /data

        # Start the IPython kernel with specified ports and settings
        CMD ["ipython", "kernel", \
        "--ip=0.0.0.0", \
        "--transport=tcp", \
        "--shell=5555", \
        "--iopub=5556", \
        "--stdin=5557", \
        "--control=5558", \
        "--hb=5559", \
        "--Session.key=19749810-8febfa748186a01da2f7b28c", \
        "--Session.signature_scheme=hmac-sha256"]
        '''

        dockerfile_legacy = 'FROM python:3.9-alpine'
        dockerfile_legacy += '\n\n'
        dockerfile_legacy += 'RUN mkdir /data'
        dockerfile_legacy += '\n\n'
        dockerfile_legacy += '# Data directory, bound as a volume to the local \'data/\' directory'
        dockerfile_legacy += '\nWORKDIR /data'

        plugin.add_option(
            "sandbox_ipython",
            type="bool",
            value=False,
            label="Sandbox (docker container)",
            description="Executes commands in sandbox (docker container). "
                        "Docker must be installed and running.",
            tab="ipython",
        )
        plugin.add_option(
            "ipython_dockerfile",
            type="textarea",
            value=dockerfile,
            label="Dockerfile for IPython kernel",
            description="Dockerfile used to build IPython kernel container image",
            tooltip="Dockerfile",
            tab="ipython",
        )
        plugin.add_option(
            "ipython_image_name",
            type="text",
            value='pygpt_ipython_kernel',
            label="Docker image name",
            tab="ipython",
        )
        plugin.add_option(
            "ipython_container_name",
            type="text",
            value='pygpt_ipython_kernel_container',
            label="Docker container name",
            tab="ipython",
        )
        plugin.add_option(
            "ipython_session_key",
            type="text",
            value='19749810-8febfa748186a01da2f7b28c',
            label="Session Key",
            tab="ipython",
        )
        plugin.add_option(
            "ipython_conn_addr",
            type="text",
            value='127.0.0.1',
            label="Connection Address",
            tab="ipython",
        )
        plugin.add_cmd(
            "ipython_execute",
            instruction="execute Python code in IPython interpreter (in current kernel) and get output. "
                        "Tip: when generating plots or other image data always print path to generated image at "
                        "the end and provide local path (prefixed with file://, not sandbox:) to the user.",
            params=[
                {
                    "name": "code",
                    "type": "str",
                    "description": "code to execute in IPython interpreter, usage of !magic commands is allowed",
                    "required": True,
                },
            ],
            enabled=True,
            description="Allows Python code execution in IPython interpreter (in current kernel)",
            tab="ipython",
        )
        """
        plugin.add_cmd(
            "ipython_execute_new",
            instruction="execute Python code in the IPython interpreter in a new kernel and get the output. Use this option only if a kernel restart is required; otherwise, use `ipython_execute` to run the code in the current session",
            params=[
                {
                    "name": "code",
                    "type": "str",
                    "description": "code to execute in IPython interpreter, usage of !magic commands is allowed",
                    "required": True,
                },
            ],
            enabled=True,
            description="Allows Python code execution in IPython interpreter (in new kernel)",
            tab="ipython",
        )
        """

        volumes_keys = {
            "enabled": "bool",
            "docker": "text",
            "host": "text",
        }
        volumes_items = [
            {
                "enabled": True,
                "docker": "/data",
                "host": "{workdir}",
            },
        ]
        ports_keys = {
            "enabled": "bool",
            "docker": "text",
            "host": "int",
        }
        ports_items = []

        plugin.add_cmd(
            "ipython_kernel_restart",
            instruction="restart IPython kernel",
            params=[],
            enabled=True,
            description="Allows to restart IPython kernel",
            tab="ipython",
        )
        plugin.add_option(
            "ipython_port_shell",
            type="int",
            value=5555,
            label="Port: shell",
            tab="ipython",
            advanced=True,
        )
        plugin.add_option(
            "ipython_port_iopub",
            type="int",
            value=5556,
            label="Port: iopub",
            tab="ipython",
            advanced=True,
        )
        plugin.add_option(
            "ipython_port_stdin",
            type="int",
            value=5557,
            label="Port: stdin",
            tab="ipython",
            advanced=True,
        )
        plugin.add_option(
            "ipython_port_control",
            type="int",
            value=5558,
            label="Port: control",
            tab="ipython",
            advanced=True,
        )
        plugin.add_option(
            "ipython_port_hb",
            type="int",
            value=5559,
            label="Port: hb",
            tab="ipython",
            advanced=True,
        )
        plugin.add_option(
            "sandbox_docker",
            type="bool",
            value=False,
            label="Sandbox (docker container)",
            description="Executes commands in sandbox (docker container). "
                        "Docker must be installed and running.",
            tab="python_legacy",
        )
        plugin.add_option(
            "python_cmd_tpl",
            type="text",
            value="python3 {filename}",
            label="Python command template",
            description="Python command template to execute, use {filename} for filename placeholder",
            tab="python_legacy",
        )
        plugin.add_option(
            "dockerfile",
            type="textarea",
            value=dockerfile_legacy,
            label="Dockerfile",
            description="Dockerfile",
            tooltip="Dockerfile",
            tab="python_legacy",
        )
        plugin.add_option(
            "image_name",
            type="text",
            value='pygpt_python_legacy',
            label="Docker image name",
            tab="python_legacy",
        )
        plugin.add_option(
            "container_name",
            type="text",
            value='pygpt_python_legacy_container',
            label="Docker container name",
            tab="python_legacy",
        )
        plugin.add_option(
            "docker_entrypoint",
            type="text",
            value='tail -f /dev/null',
            label="Docker run command",
            tab="python_legacy",
            advanced=True,
        )
        plugin.add_option(
            "docker_volumes",
            type="dict",
            value=volumes_items,
            label="Docker volumes",
            description="Docker volumes mapping",
            tooltip="Docker volumes mapping",
            keys=volumes_keys,
            tab="python_legacy",
            advanced=True,
        )
        plugin.add_option(
            "docker_ports",
            type="dict",
            value=ports_items,
            label="Docker ports",
            description="Docker ports mapping",
            tooltip="Docker ports mapping",
            keys=ports_keys,
            tab="python_legacy",
            advanced=True,
        )
        plugin.add_option(
            "attach_output",
            type="bool",
            value=True,
            label="Connect to the Python code interpreter window",
            description="Attach code input/output to the Python code interpreter window.",
            tab="general",
        )

        # commands
        plugin.add_cmd(
            "code_execute",
            instruction="save generated Python code and execute it",
            params=[
                {
                    "name": "path",
                    "type": "str",
                    "description": "path to save",
                    "default": ".interpreter.current.py",
                    "required": True,
                },
                {
                    "name": "code",
                    "type": "str",
                    "description": "code",
                    "required": True,
                },
            ],
            enabled=False,
            description="Allows Python code execution (generate and execute from file)",
            tab="python_legacy",
        )
        plugin.add_cmd(
            "code_execute_file",
            instruction="execute Python code from existing file",
            params=[
                {
                    "name": "path",
                    "type": "str",
                    "description": "file path",
                    "required": True,
                },
            ],
            enabled=False,
            description="Allows Python code execution from existing file",
            tab="python_legacy",
        )
        plugin.add_cmd(
            "code_execute_all",
            instruction="run all Python code from my interpreter",
            params=[
                {
                    "name": "code",
                    "type": "str",
                    "description": "code to append and execute",
                    "required": True,
                },
            ],
            enabled=False,
            description="Allows Python code execution (generate and execute from file)",
            tab="python_legacy",
        )
        plugin.add_cmd(
            "get_python_output",
            instruction="get output from my Python interpreter",
            params=[],
            enabled=True,
            description="Allows to get output from last executed code",
            tab="general",
        )
        plugin.add_cmd(
            "get_python_input",
            instruction="get all input code from my Python interpreter",
            params=[],
            enabled=True,
            description="Allows to get input from Python interpreter",
            tab="general",
        )
        plugin.add_cmd(
            "clear_python_output",
            instruction="clear output from my Python interpreter",
            params=[],
            enabled=True,
            description="Allows to clear output from last executed code",
            tab="general",
        )
        plugin.add_cmd(
            "render_html_output",
            instruction="send HTML/JS code to HTML built-in browser (HTML Canvas) and render it",
            params=[
                {
                    "name": "html",
                    "type": "str",
                    "description": "HTML/JS code",
                    "required": True,
                },
            ],
            enabled=True,
            description="Allows to render HTML/JS code in HTML Canvas",
            tab="html_canvas",
        )
        plugin.add_cmd(
            "get_html_output",
            instruction="get current output from HTML Canvas",
            params=[],
            enabled=True,
            description="Allows to get current output from HTML Canvas",
            tab="html_canvas",
        )


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.14 08:00:00                  #
# ================================================== #

import time
from typing import Optional

from pygpt_net.item.ctx import CtxMeta
from pygpt_net.provider.core.ctx.base import BaseProvider


class Idx:
    def __init__(self, window=None):
        """
        Context indexing helpers core

        :param window: Window instance
        """
        self.window = window

    def get_provider(self) -> BaseProvider:
        """
        Get index config storage provider instance

        :return: provider instance
        """
        return self.window.core.ctx.get_provider()

    def set_meta_indexed_by_id(
            self,
            id: int,
            doc_id: str,
            ts: int
    ) -> bool:
        """
        Update ctx meta indexed timestamp

        :param id: ctx meta ID
        :param doc_id: document ID
        :param ts: timestamp
        :return: True if updated
        """
        return self.get_provider().set_meta_indexed_by_id(id, ts)

    def update_meta_indexed_by_id(
            self,
            id: int,
            doc_id: str
    ) -> bool:
        """
        Update ctx meta indexed timestamp

        :param id: ctx meta ID
        :param doc_id: document ID
        :return: True if updated
        """
        return self.get_provider().update_meta_indexed_by_id(id)

    def update_meta_indexed_to_ts(self, ts: int) -> bool:
        """
        Update ctx meta updated timestamp

        :param ts: timestamp to update to
        :return: True if updated
        """
        return self.get_provider().update_meta_indexed_to_ts(ts)

    def clear_meta_indexed_by_id(self, id: int) -> bool:
        """
        Clear ctx meta indexed timestamp

        :param id: ctx meta ID
        :return: True if updated
        """
        return self.get_provider().clear_meta_indexed_by_id(id)

    def clear_meta_indexed_all(self) -> bool:
        """
        Clear all ctx meta indexed timestamps

        :return: True if cleared
        """
        return self.get_provider().clear_meta_indexed_all()

    def remove_meta_from_indexed(
            self,
            store: str,
            id: int,
            idx: str,
            doc_id: Optional[str] = None
    ) -> bool:
        """
        Remove ctx meta from indexed

        :param store: store name
        :param id: ctx meta ID
        :param idx: index name
        :param doc_id: document ID (optional)
        :return: True if removed
        """
        # get meta
        meta = self.window.core.ctx.get_meta_by_id(id)
        if meta is None:
            return False

        # remove from ctx index db
        self.window.core.idx.ctx.remove(store, idx, id)

        # remove document from idx storage
        if store in meta.indexes:
            if idx in list(meta.indexes[store]):
                for doc_id in meta.indexes[store][idx]:
                    self.window.core.idx.remove_doc(idx, doc_id)

        # remove index data from meta indexes
        self.remove_idx_data_from_meta(meta, store, idx)

        # update indexes in db
        self.get_provider().update_meta_indexes_by_id(id, meta)

        # clear indexed timestamp if no indexes left
        if not meta.indexes:
            self.clear_meta_indexed_by_id(id)
        return True

    def set_meta_as_indexed(
            self,
            id: int,
            idx: str,
            doc_id: str
    ) -> bool:
        """
        Set ctx meta as indexed

        :param id: ctx meta ID
        :param idx: index name
        :param doc_id: document ID
        :return: True if updated
        """
        # get meta
        meta = self.window.core.ctx.get_meta_by_id(id)
        if meta is None:
            return False

        ts = int(time.time())
        store = self.window.core.idx.get_current_store()

        # add to index db
        if not self.window.core.idx.ctx.exists(store, idx, id):
            self.window.core.idx.ctx.append(
                store_id=store,  # current store
                idx=idx,  # index name
                meta_id=id,  # ctx meta id
                doc_id=doc_id,  # document id
            )
        else:
            # update document id in index db if already indexed, by meta id
            self.window.core.idx.ctx.update(
                meta_id=id,  # ctx meta id
                doc_id=doc_id,  # document id
            )

        # update ctx meta indexed timestamp
        self.get_provider().set_meta_indexed_by_id(id, ts)
        self.window.core.ctx.update_indexed_ts_by_id(id, ts)

        # append index data to ctx meta object
        self.store_idx_data_in_meta(meta, store, idx, doc_id)

        # update ctx meta indexes data in db
        self.get_provider().update_meta_indexes_by_id(id, meta)
        return True

    def store_idx_data_in_meta(
            self,
            ctx: CtxMeta,
            store_id: str,
            idx: str,
            doc_id: str
    ):
        """
        Append index data to ctx meta object

        :param ctx: ctx meta
        :param store_id: store id
        :param idx: index name
        :param doc_id: document id
        """
        current = ctx.indexes
        if not isinstance(current, dict):
            current = {}
        if store_id not in current:
            current[store_id] = {}
        if idx not in current[store_id]:
            current[store_id][idx] = {}
        current[store_id][idx][doc_id] = int(time.time())  # key = doc_id, value = last indexed timestamp
        ctx.indexes = current  # update ctx meta object

    def remove_idx_data_from_meta(
            self,
            ctx: CtxMeta,
            store_id: str,
            idx: str,
            doc_id: Optional[str] = None
    ):
        """
        Remove index data from ctx meta object

        :param ctx: ctx meta
        :param store_id: store id
        :param idx: index name
        :param doc_id: document id
        """
        current = ctx.indexes
        if not isinstance(current, dict):
            current = {}
        if store_id not in current:
            current[store_id] = {}
        if idx not in current[store_id]:
            current[store_id][idx] = {}

        # remove doc_id from index or all docs if doc_id is None
        if doc_id is None:
            del current[store_id][idx]
        else:
            if doc_id in current[store_id][idx]:
                current[store_id][idx].remove(doc_id)
            # remove empty index
            if not current[store_id][idx]:
                del current[store_id][idx]

        # remove empty store (if no indexes left)
        if not current[store_id]:
            del current[store_id]

        ctx.indexes = current  # update ctx meta object

    def truncate_indexed(
            self,
            store_id: Optional[str] = None,
            idx: Optional[str] = None
    ) -> bool:
        """
        Truncate ctx meta indexed timestamps and statuses

        :param store_id: store id (optional)
        :param idx: index name (optional)
        :return: True if truncated
        """
        metas = self.get_provider().get_meta_indexed()
        for meta_id in metas:
            meta = metas[meta_id]
            if store_id and idx:
                if store_id in meta.indexes and idx in meta.indexes[store_id]:
                    # remove index data from meta indexes
                    self.remove_idx_data_from_meta(meta, store_id, idx)

                    # update indexes in db
                    self.get_provider().update_meta_indexes_by_id(meta.id, meta)

                    # clear indexed timestamp if no indexes left
                    if not meta.indexes:
                        self.clear_meta_indexed_by_id(meta.id)
        return True



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.11.18 19:00:00                  #
# ================================================== #

from pygpt_net.plugin.base.config import BaseConfig, BasePlugin


class Config(BaseConfig):
    def __init__(self, plugin: BasePlugin = None, *args, **kwargs):
        super(Config, self).__init__(plugin)
        self.plugin = plugin

    def from_defaults(self, plugin: BasePlugin = None):
        """
        Set default options for plugin

        :param plugin: plugin instance
        """
        plugin.add_option(
            "model_tmp_query",
            type="combo",
            value="gpt-3.5-turbo",
            label="Model for query in-memory index",
            description="Model used for query in-memory index for `query_file` command, "
                        "default: gpt-3.5-turbo",
            tooltip="Query model",
            use="models",
            tab="indexing",
        )
        plugin.add_option(
            "idx",
            type="text",
            value="base",
            label="Index to use when indexing files",
            description="ID of index to use for files indexing",
            tooltip="Index name",
            tab="indexing",
        )
        plugin.add_option(
            "use_loaders",
            type="bool",
            value=True,
            label="Use data loaders",
            description="Use data loaders from Llama-index for file reading (read_file command)",
        )
        plugin.add_option(
            "auto_index",
            type="bool",
            value=False,
            label="Auto index reading files",
            description="If enabled, every time file is read, it will be automatically indexed",
            tab="indexing",
        )
        plugin.add_option(
            "only_index",
            type="bool",
            value=False,
            label="Only index reading files",
            description="If enabled, file will be indexed without reading it",
            tab="indexing",
        )

        # commands
        plugin.add_cmd(
            "send_file",
            instruction="send file as attachment from my computer to you for analyze",
            params=[
                {
                    "name": "path",
                    "type": "str",
                    "description": "path",
                    "required": True,
                },
            ],
            enabled=True,
            description="Enable: Upload file as attachment",
        )
        plugin.add_cmd(
            "read_file",
            instruction="read data from files",
            params=[
                {
                    "name": "path",
                    "type": "list",
                    "description": "path(s) to files",
                    "required": True,
                },
            ],
            enabled=True,
            description="Enable: Read file",
        )
        plugin.add_cmd(
            "query_file",
            instruction="read, index and quick query file for additional context",
            params=[
                {
                    "name": "path",
                    "type": "str",
                    "description": "path",
                    "required": True,
                },
                {
                    "name": "query",
                    "type": "str",
                    "description": "query",
                    "required": True,
                },
            ],
            enabled=True,
            description="Enable: Query file with Llama-index",
            tab="indexing",
        )
        plugin.add_cmd(
            "save_file",
            instruction="save data to file",
            params=[
                {
                    "name": "path",
                    "type": "str",
                    "description": "path to file",
                    "required": True,
                },
                {
                    "name": "data",
                    "type": "str",
                    "description": "text data",
                    "required": True,
                },
            ],
            enabled=True,
            description="Enable: Save file",
        )
        plugin.add_cmd(
            "append_file",
            instruction="append data to file",
            params=[
                {
                    "name": "path",
                    "type": "str",
                    "description": "path",
                    "required": True,
                },
                {
                    "name": "data",
                    "type": "str",
                    "description": "data",
                    "required": True,
                },
            ],
            enabled=True,
            description="Enable: Append to file",
        )
        plugin.add_cmd(
            "delete_file",
            instruction="delete file",
            params=[
                {
                    "name": "path",
                    "type": "str",
                    "description": "path",
                    "required": True,
                },
            ],
            enabled=True,
            description="Enable: Delete file",
        )
        plugin.add_cmd(
            "list_dir",
            instruction="list files and dirs",
            params=[
                {
                    "name": "path",
                    "type": "str",
                    "description": "path",
                    "required": True,
                },
            ],
            enabled=True,
            description="Enable: List files in directory (ls)",
        )
        plugin.add_cmd(
            "tree",
            instruction="get directory tree",
            params=[
                {
                    "name": "path",
                    "type": "str",
                    "description": "path",
                    "required": True,
                },
            ],
            enabled=True,
            description="Enable: get directory tree",
        )
        plugin.add_cmd(
            "mkdir",
            instruction="create directory",
            params=[
                {
                    "name": "path",
                    "type": "str",
                    "description": "path",
                    "required": True,
                },
            ],
            enabled=True,
            description="Enable: Directory creation (mkdir)",
        )
        plugin.add_cmd(
            "download_file",
            instruction="download file",
            params=[
                {
                    "name": "src",
                    "type": "str",
                    "description": "source URL",
                    "required": True,
                },
                {
                    "name": "dst",
                    "type": "str",
                    "description": "path",
                    "required": True,
                },
            ],
            enabled=True,
            description="Enable: Downloading files",
        )
        plugin.add_cmd(
            "rmdir",
            instruction="remove directory",
            params=[
                {
                    "name": "path",
                    "type": "str",
                    "description": "path",
                    "required": True,
                },
            ],
            enabled=True,
            description="Enable: Removing directories",
        )
        plugin.add_cmd(
            "copy_file",
            instruction="copy file",
            params=[
                {
                    "name": "src",
                    "type": "str",
                    "description": "path",
                    "required": True,
                },
                {
                    "name": "dst",
                    "type": "str",
                    "description": "path",
                    "required": True,
                },
            ],
            enabled=True,
            description="Enable: Copying files",
        )
        plugin.add_cmd(
            "copy_dir",
            instruction="recursive copy directory",
            params=[
                {
                    "name": "src",
                    "type": "str",
                    "description": "path",
                    "required": True,
                },
                {
                    "name": "dst",
                    "type": "str",
                    "description": "path",
                    "required": True,
                },
            ],
            enabled=True,
            description="Enable: Copying directories (recursive)",
        )
        plugin.add_cmd(
            "move",
            instruction="move file or directory",
            params=[
                {
                    "name": "src",
                    "type": "str",
                    "description": "path",
                    "required": True,
                },
                {
                    "name": "dst",
                    "type": "str",
                    "description": "path",
                    "required": True,
                },
            ],
            enabled=True,
            description="Enable: Move files and directories (rename)",
        )
        plugin.add_cmd(
            "is_dir",
            instruction="check if path is directory",
            params=[
                {
                    "name": "path",
                    "type": "str",
                    "description": "path",
                    "required": True,
                },
            ],
            enabled=True,
            description="Enable: Check if path is directory",
        )
        plugin.add_cmd(
            "is_file",
            instruction="check if path is file",
            params=[
                {
                    "name": "path",
                    "type": "str",
                    "description": "path",
                    "required": True,
                },
            ],
            enabled=True,
            description="Enable: Check if path is file",
        )
        plugin.add_cmd(
            "file_exists",
            instruction="check if file or directory exists",
            params=[
                {
                    "name": "path",
                    "type": "str",
                    "description": "path",
                    "required": True,
                },
            ],
            enabled=True,
            description="Enable: Check if file or directory exists",
        )
        plugin.add_cmd(
            "file_size",
            instruction="get file size",
            params=[
                {
                    "name": "path",
                    "type": "str",
                    "description": "path",
                    "required": True,
                },
            ],
            enabled=True,
            description="Enable: Get file size",
        )
        plugin.add_cmd(
            "file_info",
            instruction="get file info",
            params=[
                {
                    "name": "path",
                    "type": "str",
                    "description": "path",
                    "required": True,
                },
            ],
            enabled=True,
            description="Enable: Get file info",
        )
        plugin.add_cmd(
            "cwd",
            instruction="get current working directory (abs path)",
            params=[],
            enabled=True,
            description="Enable: Get current working directory (cwd)",
        )
        plugin.add_cmd(
            "file_index",
            instruction="index (embed as vectors in vector DB) file or directory",
            params=[
                {
                    "name": "path",
                    "type": "str",
                    "description": "path",
                    "required": True,
                },
            ],
            enabled=True,
            description="If enabled, model will be able to index file or directory using Llama-index",
        )
        plugin.add_cmd(
            "find",
            instruction="find file or directory, use empty path to search in current dir",
            params=[
                {
                    "name": "pattern",
                    "type": "str",
                    "description": "name pattern",
                    "required": True,
                },
                {
                    "name": "path",
                    "type": "str",
                    "description": "search directory",
                    "required": True,
                },
                {
                    "name": "recursive",
                    "type": "bool",
                    "description": "recursive search",
                    "required": True,
                },
            ],
            enabled=True,
            description="Enable: Find file or directory",
        )


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.14 08:00:00                  #
# ================================================== #

from typing import Dict, List

from PySide6.QtGui import QAction, QIcon
from PySide6.QtWidgets import QTabWidget, QMenu

from pygpt_net.core.tabs.tab import Tab
from pygpt_net.utils import trans


class Tools:
    def __init__(self, window=None):
        """
        Tools controller

        :param window: Window instance
        """
        self.window = window
        self.tab_tools = {
            'tools.files': ['files', 'folder_filled', Tab.TAB_FILES],
            'tools.calendar': ['calendar', 'calendar', Tab.TAB_TOOL_CALENDAR],
            'tools.notepad': ['notepad', 'paste', Tab.TAB_NOTEPAD],
            'tools.painter': ['painter', 'brush', Tab.TAB_TOOL_PAINTER],
        }

    def setup(self):
        """Setup tools"""
        pass

    def reload(self):
        """Reload tools"""
        pass

    def open_tab(self, type: int):
        """
        Open first tab by type

        :param type: tab type
        """
        idx = self.window.core.tabs.get_min_idx_by_type(type)
        if idx is not None:
            self.window.controller.ui.tabs.switch_tab_by_idx(idx)

    def append_tab_menu(
            self,
            parent: QTabWidget,
            menu: QMenu,
            idx: int,
            column_idx: int
    ) -> QMenu:
        """
        Append tab menu

        :param parent: parent widget
        :param menu: menu
        :param idx: tab index
        :param column_idx: column index
        :return: tab add submenu
        """
        submenu = menu.addMenu(QIcon(":/icons/add.svg"), trans("action.tab.add.tool"))
        tools = self.window.tools.get_all()
        for id in tools:
            tool = tools[id]
            if not tool.has_tab:
                continue
            icon = tool.tab_icon
            title = trans(tool.tab_title)
            action = QAction(QIcon(icon), title, parent)
            action.triggered.connect(
                lambda idx=idx, column_idx=column_idx, id=id: parent.add_tab(idx, column_idx, Tab.TAB_TOOL, id)
            )
            submenu.addAction(action)
        return submenu

    def get_tab_tools(self) -> Dict[str, List[str]]:
        """
        Get tab tools

        :return: tab tools
        """
        return self.tab_tools



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.11.21 20:00:00                  #
# ================================================== #

from .common import Common
from .experts import Experts
from .legacy import Legacy
from .llama import Llama

class Agent:
    def __init__(self, window=None):
        """
        Agent controller

        :param window: Window instance
        """
        self.window = window
        self.common = Common(window)
        self.experts = Experts(window)
        self.llama = Llama(window)
        self.legacy = Legacy(window)

    def setup(self):
        """Setup agent controller"""
        self.legacy.setup()
        self.llama.setup()

    def reload(self):
        """Reload agent toolbox options"""
        self.legacy.reload()
        self.llama.reload()

    def stop(self):
        """Force stop all agents"""
        self.legacy.on_stop()
        self.llama.on_stop()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.11.18 19:00:00                  #
# ================================================== #

from pygpt_net.plugin.base.config import BaseConfig, BasePlugin


class Config(BaseConfig):
    def __init__(self, plugin: BasePlugin = None, *args, **kwargs):
        super(Config, self).__init__(plugin)
        self.plugin = plugin

    def from_defaults(self, plugin: BasePlugin = None):
        """
        Set default options for plugin

        :param plugin: plugin instance
        """
        prompt = 'IMAGE GENERATION: Whenever I provide a basic idea or concept for an image, such as \'a picture of ' \
                 'mountains\', I want you to ALWAYS translate it into English and expand and elaborate on this idea. ' \
                 'Use your  knowledge and creativity to add details that would make the image more vivid and ' \
                 'interesting. This could include specifying the time of day, weather conditions, surrounding ' \
                 'environment, and any additional elements that could enhance the scene. Your goal is to create a ' \
                 'detailed and descriptive prompt that provides DALL-E  with enough information to generate a rich ' \
                 'and visually appealing image. Remember to maintain the original  intent of my request while ' \
                 'enriching the description with your imaginative details. HOW TO START IMAGE GENERATION: to start ' \
                 'image generation return to me prepared prompt in JSON format, all in one line,  using following ' \
                 'syntax: ~###~{"cmd": "image", "params": {"query": "your query here"}}~###~. Use ONLY this syntax ' \
                 'and remember to surround JSON string with ~###~. DO NOT use any other syntax. Use English in the ' \
                 'generated JSON command, but conduct all the remaining parts of the discussion with me in the ' \
                 'language in which I am speaking to you. The image will be generated on my machine  immediately ' \
                 'after the command is issued, allowing us to discuss the photo once it has been created.  Please ' \
                 'engage with me about the photo itself, not only by giving the generate command. '
        prompt_func = 'IMAGE GENERATION: Whenever I provide a basic idea or concept for an image, such as \'give me a picture of ' \
                      'mountains\', I want you to translate it into English and expand and elaborate on this idea. ' \
                      'Use your knowledge and creativity to add details that would make the image more vivid and ' \
                      'interesting. This could include specifying the time of day, weather conditions, surrounding ' \
                      'environment, and any additional elements that could enhance the scene. Your goal is to create a ' \
                      'detailed and descriptive prompt (query) that provides DALL-E with enough information to generate a rich ' \
                      'and visually appealing image. Use this command to start image generation. Use English in the image query.  ' \
                      'The image will be generated on my machine immediately after the command is issued, allowing us to ' \
                      'discuss the photo once it has been created. Please engage with me about the photo itself, not only by giving the generate command. '
        plugin.add_cmd(
            "image",
            instruction=prompt_func,
            params=[
                {
                    "name": "query",
                    "type": "str",
                    "description": "image generation query (prompt for DALL-E)",
                    "required": True,
                },
            ],
            enabled=True,
            description="If enabled, DALL-E 3 image generation is available in chat.",
        )
        plugin.add_option(
            "prompt",
            type="textarea",
            value=prompt,
            label="Prompt",
            description="Prompt used for generating a query for DALL-E in background.",
            tooltip="Prompt",
            advanced=False,
        )


# !/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.14 22:00:00                  #
# ================================================== #

from typing import Dict, Optional

from PySide6.QtGui import QAction

from pygpt_net.ui.widget.dialog.base import BaseDialog
from .base import BaseTool

class Tools:
    def __init__(self, window=None):
        """
        Tools manager

        :param window: Window instance
        """
        self.window = window
        self.tools = {}
        self.initialized = False

    def register(self, tool: BaseTool):
        """
        Register tool

        :param tool: Tool instance
        """
        self.tools[tool.id] = tool
        self.tools[tool.id].attach(self.window)

    def get(self, id: str) -> BaseTool:
        """
        Get tool instance by ID

        :param id: tool ID
        :return tool instance
        """
        if id in self.tools:
            return self.tools[id]

    def get_all(self) -> Dict[str, BaseTool]:
        """
        Get all tools

        :return: dict with tools
        """
        return self.tools

    def setup(self):
        """Setup tools"""
        self.setup_dialogs()
        for id in self.tools:
            self.tools[id].setup()
        self.setup_theme()
        self.initialized = True

    def post_setup(self):
        """Post-setup, after plugins are loaded"""
        for id in self.tools:
            self.tools[id].post_setup()

    def on_update(self):
        """On app main loop update"""
        for id in self.tools:
            self.tools[id].on_update()

    def on_post_update(self):
        """On app main loop post update"""
        for id in self.tools:
            self.tools[id].on_post_update()

    def on_exit(self):
        """On app exit"""
        for id in self.tools:
            self.tools[id].on_exit()

    def on_reload(self):
        """On app profile reload"""
        for id in self.tools:
            self.tools[id].on_reload()

    def setup_menu_actions(self) -> Dict[str, QAction]:
        """
        Setup Tools menu actions

        :return: dict with menu actions
        """
        actions = {}
        for id in self.tools:
            tool_actions = self.tools[id].setup_menu()
            if tool_actions and isinstance(tool_actions, dict):
                for id in tool_actions:
                    key = "tools." + id
                    actions[key] = tool_actions[id]
        return actions

    def setup_dialogs(self):
        """Setup dialogs"""
        for id in self.tools:
            self.tools[id].setup_dialogs()

    def setup_theme(self):
        """Setup theme"""
        if not self.initialized:
            return
        for id in self.tools:
            self.tools[id].setup_theme()

    def get_instance(
            self,
            type_id: str,
            dialog_id: Optional[str] = None
    ) -> Optional[BaseDialog]:
        """
        Spawn and return dialog instance

        :param type_id: dialog instance type ID
        :param dialog_id: dialog instance ID
        :return: BaseDialog instance or None
        """
        for id in self.tools:
            instance = self.tools[id].get_instance(type_id, dialog_id)
            if instance is not None:
                return instance

    def get_lang_mappings(self) -> Dict[str, dict]:
        """
        Get language mappings

        :return: dict with language mappings
        """
        mappings = {}
        for id in self.tools:
            tool_mappings = self.tools[id].get_lang_mappings()
            if tool_mappings and isinstance(tool_mappings, dict):
                for key in tool_mappings:
                    if key not in mappings:
                        mappings[key] = tool_mappings[key]
                    else:
                        mappings[key].update(tool_mappings[key])
        return mappings
        



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.09 03:00:00                  #
# ================================================== #

from PySide6.QtCore import Qt, QObject, QEvent
from PySide6.QtWidgets import QTabWidget, QWidget, QVBoxLayout, QSplitter, QSizePolicy

from pygpt_net.ui.widget.tabs.output import OutputTabs


class OutputColumn(QWidget):
    def __init__(self, window=None):
        """
        Output column

        :param window: window instance
        """
        super(OutputColumn, self).__init__(window)
        self.window = window
        self.idx = -1
        self.tabs = OutputTabs(self.window, column=self)
        self.layout = QVBoxLayout()
        self.layout.addWidget(self.tabs)
        self.layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(self.layout)
        self.filter = FocusEventFilter(self, self.on_focus)
        self.installEventFilter(self.filter)
        self.setFocusPolicy(Qt.StrongFocus)

    def on_focus(self, widget):
        """
        On widget clicked

        :param widget: widget
        """
        self.window.controller.ui.tabs.on_column_focus(self.idx)
        widget.setFocus()

    def set_idx(self, idx: int):
        """
        Set index

        :param idx: int
        """
        self.idx = idx

    def get_idx(self) -> int:
        """
        Get index

        :return: int
        """
        return self.idx

    def set_tabs(self, tabs: QTabWidget):
        """
        Set tabs widget

        :param tabs: QTabWidget
        """
        self.tabs = tabs

    def get_tabs(self) -> OutputTabs:
        """
        Get tabs

        :return: OutputTabs
        """
        return self.tabs


class OutputLayout(QWidget):
    def __init__(self, window=None):
        """
        Output layout

        :param window: window instance
        """
        super(OutputLayout, self).__init__(window)
        self.window = window
        self.columns = []

        column1 = OutputColumn(self.window)
        column2 = OutputColumn(self.window)
        self.add_column(column1)
        self.add_column(column2)

        self.splitter = QSplitter(Qt.Horizontal)
        self.splitter.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        for column in self.columns:
            self.splitter.addWidget(column)

        self.window.ui.splitters['columns'] = self.splitter

        self.layout = QVBoxLayout()
        self.layout.addWidget(self.splitter, stretch=1)
        self.layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(self.layout)

    def get_next_idx(self) -> int:
        """
        Get next index

        :return: int
        """
        return len(self.columns)

    def add_column(self, column: OutputColumn):
        """
        Add column

        :param column: OutputColumn
        """
        idx = self.get_next_idx()
        column.set_idx(idx)
        self.columns.append(column)

    def get_tabs_by_idx(self, idx: int) -> OutputTabs:
        """
        Get tabs by column index

        :param idx: int
        :return: OutputTabs
        """
        for column in self.columns:
            if column.idx == idx:
                return column.tabs
        return None

    def get_active_tabs(self) -> OutputTabs:
        """
        Get active tabs

        :return: OutputTabs
        """
        current = self.window.controller.ui.tabs.get_current_column_idx()
        for column in self.columns:
            if column.idx == current:
                return column.tabs

    def get_column_by_idx(self, idx: int) -> OutputColumn:
        """
        Get column by index

        :param idx: int
        :return: OutputColumn
        """
        for column in self.columns:
            if column.idx == idx:
                return column
        return None

    def get_active_column(self) -> OutputColumn:
        """
        Get active column

        :return: OutputColumn
        """
        current = self.window.controller.ui.tabs.get_current_column_idx()
        for column in self.columns:
            if column.idx == current:
                return column

class FocusEventFilter(QObject):
    def __init__(self, column, callback):
        """
        Column event filter

        :param column: parent column
        :param callback: callback
        """
        super().__init__()
        self.column = column
        self.callback = callback

    def eventFilter(self, obj, event):
        """
        Click event filter

        :param obj: object
        :param event: event
        """
        if event.type() == QEvent.MouseButtonPress or event.type() == QEvent.FocusIn:
            widget = obj
            if widget is not None:
                self.callback(widget)
            return False
        return False
